from datetime import datetime
from typing import Optional

from ..base_object import BaseDiscordObject
from .. import snowflake
from ... import objects


class TypingStart(BaseDiscordObject):
    '''Event given when a user starts to type.

    Attributes:
        channel_id (Snowflake): ID of the typing event.
        guild_id (Snowflake): Guild ID of the typing event.
        user_id (Snowflake): user_id of the user which started to type.
        timestamp (datetime): Datetime the user started typing.
        member (Member): Member that started typing.
    '''

    channel_id: 'objects.Snowflake' = None  # type: ignore
    guild_id: 'Optional[objects.Snowflake]' = None  # type: ignore
    user_id: 'objects.Snowflake' = None  # type: ignore
    timestamp: datetime = None  # type: ignore
    member: 'Optional[objects.Member]' = None  # type: ignore

    def from_dict(self, data: dict) -> 'TypingStart':
        '''Parse a TypingStart from an API compliant dict.'''
        if 'channel_id' in data:
            self.channel_id = snowflake.Snowflake(data['channel_id'])
        if 'guild_id' in data:
            self.guild_id = snowflake.Snowflake(data['guild_id'])
        if 'user_id' in data:
            self.user_id = snowflake.Snowflake(data['user_id'])
        if 'timestamp' in data:
            self.timestamp = datetime.fromtimestamp(data['timestamp'])
        if 'member' in data:
            self.member = objects.Member().from_dict(data['member'])
        return self



import datetime
from typing import Union


class Snowflake:
    '''Discord specific UUID like object. Generally interchangable with a string contained the same sequence of characters.'''

    DISCORD_EPOCH = 1420070400000

    def __init__(self, identifier: Union[int, str, 'Snowflake'] = None):
        '''Initialize the Snowflake.

        Arguments:
            identifier (int, str, Snowflake): Identifier to base off of.
        '''
        self.identifier: int

        if type(identifier) is int:
            self.identifier = identifier
        elif type(identifier) is str:
            self.identifier = int(identifier)
        elif type(identifier) is Snowflake:
            self.identifier = identifier.identifier
        else:
            self.generate()

        assert type(self.identifier) is int

    def __str__(self):
        '''Return string representation.'''
        return str(self.identifier)

    def __repr__(self):
        '''Return string representation.'''
        return self.__str__()

    def __eq__(self, other):
        '''Determine if other is equivalent to this Snowflake..'''
        if type(other) in [str, int]:
            other = Snowflake(other)
        elif type(other) is Snowflake:
            pass
        else:
            return NotImplemented

        return self.identifier == other.identifier

    def __hash__(self):
        '''Hash.'''
        return hash(self.identifier)

    @property
    def timestamp(self):
        '''The timestamp of the Snowflake.'''
        assert type(self.identifier) is int

        return ((self.identifier >> 22) + 1420070400000) / 1000

    @property
    def worker_id(self):
        '''The worker ID that generated the Snowflake.'''
        assert type(self.identifier) is int

        return (self.identifier & 0x3E0000) >> 17

    @property
    def process_id(self):
        '''The process ID of the Snowflake.'''
        assert type(self.identifier) is int

        return (self.identifier & 0x1F000) >> 12

    @property
    def increment(self):
        '''The increment sequence count of the Snowflake.'''
        assert type(self.identifier) is int

        return self.identifier & 0xFFF

    def generate(self):
        '''Generate a fake Snowflake locally.'''
        self.identifier = (int(datetime.datetime.now().timestamp() * 1000) - 1420070400000) << 22



'''Commands that can be registered with the API.'''

import abc
import builtins
import copy
import re
from typing import Optional, Union, List, Dict

from ...client import api

from ..base_object import BaseDiscordObject

from .. import snowflake, guild as ext_guild

from . import enumerations


class Command(BaseDiscordObject):
    '''Command root used to generate new commands.

    Attributes:
        COMMAND_TYPE (COMMAND_TYPE): Helper pointer to the COMMAND_TYPE enumeration.
        COMMAND_OPTION (COMMAND_OPTION): Helper pointer to the COMMAND_OPTION enumeration.
        id (Snowflake): Unique ID of the command.
        type (Optional[COMMAND_TYPE]): Type of the command.
        application_id: (Snowflake): Unique id of the parent application.

    '''

    COMMAND_TYPE = enumerations.COMMAND_TYPE
    COMMAND_OPTION = enumerations.COMMAND_OPTION

    id: 'snowflake.Snowflake'                        # unique id of the command  all
    type: Optional['enumerations.COMMAND_TYPE']      # one of application command type  the type of command, defaults 1 if not set  all
    application_id: Optional['snowflake.Snowflake']  # unique id of the parent application  all
    guild_id: Optional['snowflake.Snowflake']        # guild id of the command, if not global  all
    name: str                                        # 1-32 character name  all
    description: str                                 # 1-100 character description for CHAT_INPUT commands, empty string for USER and MESSAGE commands  all
    options: Optional[List['CommandOptionsBase']]    # array of application command option  the parameters for the command, max 25  CHAT_INPUT
    default_permission: bool                         # boolean (default true)  whether the command is enabled by default when the app is added to a guild  all
    version: Optional['snowflake.Snowflake']         # autoincrementing version identifier updated during substantial record changes  all

    def __str__(self):
        '''Return string representation.'''
        fields = list()

        if hasattr(self, 'id'):
            fields.append(f'id={self.id}')
        if hasattr(self, 'type') and type(self.type) is enumerations.COMPONENT_TYPES:
            fields.append(f'type={self.type.name}')
        if hasattr(self, 'name'):
            fields.append(f'name=\'{self.name}\'')
        if hasattr(self, 'description'):
            fields.append(f'description=\'{self.description}\'')
        if hasattr(self, 'version'):
            fields.append(f'version={self.version}')

        return f'Command({", ".join(fields)})'

    def generate(self,
                 name: str,
                 description: str,
                 type: 'enumerations.COMMAND_TYPE',
                 options: Optional[List['CommandOptionsBase']] = None,
                 default_permission: bool = True,
                 ):
        '''Helper function to generate a new command bound for registration.

        Arguments:
            name (str): Name of the command, will be referenced when registering the command with discord as well as locally.
            description (str): Human readable description. For USER and MESSAGE types, this *must* be a zero length string.
            type (COMMAND_TYPE): Either CHAT_INPUT for a slash command, or USER/MESSAGE for direct object interaction.
            options ([CommandOptionsBase]): DEPRECATED[Removed by 0.6.0] List of valid Command Options. Will be copied.
            default_permission (bool): Whether the command is enabled by default when the app is added to a guild.
        '''
        self.name = name
        self.description = description
        self.default_permission = default_permission
        self.type = type

        if builtins.type(options) is list:
            self.options = copy.deepcopy(options)
        elif options is None and hasattr(self, 'options'):
            del self.options

    def add_option_sub_command(self,
                               name: str,
                               description: str,
                               ) -> 'CommandOptionSubCommand':
        '''Add SubCommand to the command.'''
        new_option = CommandOptionSubCommand()
        new_option.type = enumerations.COMMAND_OPTION.SUB_COMMAND
        new_option.name = name
        new_option.description = description

        if not hasattr(self, 'options'):
            self.options = list()

        assert type(self.options) is list
        self.options.append(new_option)

        return new_option

    def add_option_sub_command_group(self,
                                     name: str,
                                     description: str,
                                     ) -> 'CommandOptionSubCommandGroup':
        '''Add SubCommandGroup to the command.'''
        new_option = CommandOptionSubCommandGroup()
        new_option.type = enumerations.COMMAND_OPTION.SUB_COMMAND_GROUP
        new_option.name = name
        new_option.description = description

        if not hasattr(self, 'options'):
            self.options = list()

        assert type(self.options) is list
        self.options.append(new_option)

        return new_option

    def add_option_typed(self,
                         type: 'enumerations.COMMAND_OPTION',
                         name: str,
                         description: str,
                         required: bool = True,
                         choices: Optional[List['CommandOptionChoiceStructure']] = None,
                         ) -> 'CommandOptions':
        '''Add option to the command.'''
        new_option = CommandOptions()  # type: ignore
        new_option.type = type
        new_option.name = name
        new_option.description = description
        new_option.required = required
        if choices is not None:
            new_option.choices = copy.deepcopy(choices)

        if not hasattr(self, 'options') or self.options is None:
            self.options = list()
        assert builtins.type(self.options) is list
        self.options.append(new_option)

        return new_option

    def clear_options(self):
        '''Remove all options.'''
        if hasattr(self, 'options'):
            del self.options

    async def register_to_guild(self, guild: 'Union[ext_guild.Guild, snowflake.Snowflake]') -> dict:
        '''Register a Command to a specific guild discord scope.

        Note that discord will limit you to 200 of these calls per bot per guild per day.
        '''
        if isinstance(guild, snowflake.Snowflake):
            guild_id = guild
        elif isinstance(guild, ext_guild.Guild):
            guild_id = guild.id
        return await api.API.create_guild_application_command(guild_id, self.to_dict())

    async def register_globally(self) -> dict:
        '''Register a Command to the global discord scope.

        Note that discord will limit you to 200 of these calls per bot per day.
        '''
        return await api.API.create_global_application_command(self.to_dict())

    def from_dict(self, data: dict) -> 'Command':
        '''Parse a Command from an API compliant dict.'''
        if 'id' in data:
            self.id = snowflake.Snowflake(data['id'])
        if 'type' in data:
            self.type = enumerations.COMMAND_TYPE(data['type'])
        if 'application_id' in data:
            self.application_id = data['application_id']
        if 'guild_id' in data:
            self.guild_id = data['guild_id']
        self.name = data['name']
        self.description = data['description']
        if 'options' in data:
            # TODO: Loop through and process the options.
            # self.options
            pass
        if 'default_permission' in data:
            self.default_permission = data['default_permission']
        if 'version' in data:
            self.version = data['version']

        return self

    def to_dict(self) -> dict:
        '''Convert object to dictionary suitable for API or other generic useage.'''
        ret_dict: Dict[str, Union[object, list]] = dict()
        if hasattr(self, 'id'):
            ret_dict['id'] = str(self.id)
        if hasattr(self, 'type') and self.type is not None:
            if type(self.type) is int:
                ret_dict['type'] = self.type
            else:
                ret_dict['type'] = self.type.value
        if hasattr(self, 'application_id'):
            ret_dict['application_id'] = str(self.application_id)
        if hasattr(self, 'guild_id'):
            ret_dict['guild_id'] = str(self.guild_id)
        ret_dict['name'] = self.name
        ret_dict['description'] = self.description
        if hasattr(self, 'options') and self.options is not None:
            ret_dict['options'] = list()
            assert type(ret_dict['options']) is list  # This has got to be the stupidest assert ever, but mypy wants it!
            for option in self.options:
                ret_dict['options'].append(option.to_dict())
        if hasattr(self, 'default_permission'):
            ret_dict['default_permission'] = self.default_permission
        if hasattr(self, 'version'):
            ret_dict['version'] = str(self.version)

        return ret_dict

    def validate(self):  # noqa: C901
        '''Validate object is prepared for dispatch to discord.'''
        regex = r'^[\w-]{1,32}$'
        if not re.match(regex, self.name):
            raise ValueError(f'Name of \'{self.name}\' does not match \'{regex}\'.')

        if self.name != self.name.lower():
            raise ValueError('Command name must use lower case version of all characters.')

        if len(self.description) > 100:
            raise ValueError('Description cannot exceed 100 characters.')

        if self.type in [enumerations.COMMAND_TYPE.USER, enumerations.COMMAND_TYPE.MESSAGE]:
            if len(self.description):
                raise ValueError(f'Descriptions are not allowed for type {self.type.name} commands.')
        elif self.type == enumerations.COMMAND_TYPE.CHAT_INPUT:
            if not len(self.description):
                raise ValueError(f'Descriptions are mandatory {self.type.name} commands.')

        if type(self.type) is not enumerations.COMMAND_TYPE:
            raise TypeError(f'Type is of incorrect type {type(self.type)}, should be {enumerations.COMMAND_TYPE}.')

        if hasattr(self, 'options') and type(self.options) is list:

            if len(self.options) and self.type in [enumerations.COMMAND_TYPE.USER, enumerations.COMMAND_TYPE.MESSAGE]:
                raise ValueError('Context menu commands cannot have options.')

            if len(self.options) > 25:
                raise OverflowError(f'You cannot have {len(self.options)} options, limited to max of 25.')

            for option in self.options:
                if not isinstance(option, CommandOptionsBase):
                    raise TypeError(f'Found option of type {type(option)}, must be {CommandOptions}')
                option.validate()

        if type(self.default_permission) is not bool:
            raise TypeError(f'Default Permission is of incorrect type {type(self.default_permission)}, should be {type(True)}.')

        assert self.total_characters() <= 4000,\
            f'Found {self.total_characters():d} characters in name, description, and value fields. Max is 4,000.'

    def total_characters(self) -> int:
        '''Get the total characters in the name, description, and value.'''
        total_characters = 0
        if hasattr(self, 'name') and type(self.name) is str:
            total_characters += len(self.name)
        if hasattr(self, 'description') and type(self.description) is str:
            total_characters += len(self.description)
        if hasattr(self, 'options') and type(self.options) is list:
            for option in self.options:
                total_characters += option.total_characters()
        return total_characters


class CommandOptionsBase(BaseDiscordObject, abc.ABC):
    '''Abstract base for options.'''

    # Handy shortcut
    COMMAND_OPTION = enumerations.COMMAND_OPTION

    type: 'enumerations.COMMAND_OPTION'                       # one of application command option type the type of option
    name: str                                                 # string 1-32 character name
    description: str                                          # string 1-100 character description
    required: bool                                            # boolean if the parameter is required or optional--default false
    choices: Optional[List['CommandOptionChoiceStructure']]   # array of application command option choice choices for STRING, INTEGER, and
    # NUMBER types for the user to pick from, max 25
    options: Optional[List['CommandOptions']]                 # array of application command option if the option is a subcommand or subcommand group type,
    # this nested options will be the parameters

    def from_dict(self, data: dict) -> 'CommandOptionsBase':
        '''Parse a CommandOptionsBase from an API compliant dict.'''
        if 'type' in data:
            self.type = enumerations.COMMAND_OPTION(data['type'])
        self.name = data['name']
        self.description = data['description']
        if 'required' in data:
            self.required = data['required']
        if 'choices' in data:
            # TODO: Loop through and process the choices.
            # self.choices
            pass
        if 'options' in data:
            # TODO: Loop through and process the options.
            # self.choices
            pass

        return self

    def to_dict(self) -> dict:
        '''Convert object to dictionary suitable for API or other generic useage.'''
        ret_dict: Dict[str, Union[object, list]] = dict()
        if hasattr(self, 'type'):
            ret_dict['type'] = self.type.value
        ret_dict['name'] = self.name
        ret_dict['description'] = self.description
        if hasattr(self, 'required'):
            ret_dict['required'] = self.required
        if hasattr(self, 'choices'):
            ret_dict['choices'] = list()
            assert type(ret_dict['choices']) is list  # This has got to be the stupidest assert ever, but mypy wants it!
            assert type(self.choices) is list
            for choice in self.choices:
                ret_dict['choices'].append(choice.to_dict())
        if hasattr(self, 'options'):
            ret_dict['options'] = list()
            assert type(ret_dict['options']) is list  # This has got to be the stupidest assert ever, but mypy wants it!
            assert type(self.options) is list
            for option in self.options:
                ret_dict['options'].append(option.to_dict())

        return ret_dict

    def validate(self):
        '''Validate object is prepared for dispatch to discord.'''
        regex = r'^[\w-]{1,32}$'
        if not re.match(regex, self.name):
            raise ValueError(f'Name of \'{self.name}\' does not match \'{regex}\'.')

        if self.name != self.name.lower():
            raise ValueError('Command name must use lower case version of all characters.')

        if hasattr(self, 'choices') and self.choices is not None:
            for choice in self.choices:
                choice.validate()

        if hasattr(self, 'options') and self.options is not None:
            for option in self.options:
                option.validate()

    def total_characters(self) -> int:
        '''Get the total characters in the name, description, and value.'''
        total_characters = 0
        if hasattr(self, 'name') and type(self.name) is str:
            total_characters += len(self.name)
        if hasattr(self, 'description') and type(self.description) is str:
            total_characters += len(self.description)

        if hasattr(self, 'choices') and type(self.choices) is list:
            for choice in self.choices:
                total_characters += choice.total_characters()
        if hasattr(self, 'options') and type(self.options) is list:
            for option in self.options:
                total_characters += option.total_characters()
        return total_characters


class CommandOptions(CommandOptionsBase):
    '''Options for a Command.'''

    def add_choice(self,
                   name: str,
                   value: Union[str, int, float],
                   ) -> 'CommandOptionChoiceStructure':
        '''Append a choice to the given set of options and return it. Will start a list if needed.'''
        new_choice = CommandOptionChoiceStructure()
        new_choice.name = name
        new_choice.value = value
        if not hasattr(self, 'choices') or self.choices is None:
            self.choices = list()
        self.choices.append(new_choice)

        return new_choice

    def validate(self):
        '''Validate object is prepared for dispatch to discord.'''
        return super().validate()


class CommandOptionSubCommandGroup(CommandOptionsBase):
    '''Intermediate object to group SubCommands together.'''
    # Same thing, let's just take it!
    add_option_sub_command = Command.add_option_sub_command


class CommandOptionSubCommand(CommandOptionsBase):
    '''Commands which live under a Command, or a Sub Command Group.'''

    COMMAND_OPTION = enumerations.COMMAND_OPTION

    add_option_typed = Command.add_option_typed


class CommandOptionChoiceStructure(BaseDiscordObject):
    '''Specific option for a command.'''

    def __init__(self):
        '''Initalize a CommandOptionChoiceStructure.'''
        self.name: str
        self.value: Union[str, int, float]

    def from_dict(self, data: dict) -> 'CommandOptionChoiceStructure':
        '''Parse a CommandOptionChoiceStructure from an API compliant dict.'''
        self.name = data['name']
        self.value = data['value']
        if type(self.value) not in [str, int, float]:
            raise ValueError(f'Value must be in [str, int, float], got {type(self.value)}')

        return self

    def to_dict(self) -> dict:
        '''Convert object to dictionary suitable for API or other generic useage.'''
        ret_dict: Dict[str, object] = dict()
        ret_dict['name'] = self.name
        ret_dict['value'] = self.value

        return ret_dict

    def validate(self):
        '''Validate object is prepared for dispatch to discord.'''
        regex = r'^[\w-]{1,32}$'
        if not re.match(regex, self.name):
            raise ValueError(f'Name of \'{self.name}\' does not match \'{regex}\'.')

        if len(self.name) < 1 or len(self.name) > 100:
            raise ValueError(f'Length of name is {len(self.name)}, but be [1-100].')

    def total_characters(self) -> int:
        '''Get the total characters in the name, description, and value.'''
        total_characters = 0
        if hasattr(self, 'name') and type(self.name) is str:
            total_characters += len(self.name)
        if hasattr(self, 'value') and type(self.value) is str:
            total_characters += len(self.value)
        return total_characters



import warnings
import inspect
import functools
import httpx
from cachetools import TTLCache
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Callable, Dict, Optional, Union, Any, Tuple

from ..objects import interactions, snowflake, guild as ext_guild
from ..utilities import Log
from ..client import api


@dataclass
class CallbackData:
    '''Callback dataclass.'''
    custom_id: str
    callback: Callable
    unlimited: bool


class CommandHandler:
    '''Manage commands for the user.'''
    _log = Log()

    registered_commands: 'Dict[snowflake.Snowflake, Callable]' = dict()

    global_lookup: 'Dict[str, Callable]' = dict()

    guild_lookup: 'Dict[Tuple[str, Optional[snowflake.Snowflake]], Callable]' = dict()

    registered_custom_ids: 'TTLCache[str, CallbackData]' = TTLCache(
        maxsize=float('inf'),
        ttl=timedelta(minutes=15),  # type: ignore
        timer=datetime.now,  # type: ignore
    )

    @classmethod
    def decorate_global_callback(cls):
        '''WIP Decorator to register against a target interaction.'''
        pass

    @classmethod
    def decorate_guild_callback(cls):
        '''WIP Decorator to register against a target interaction.'''
        pass

    @classmethod
    def register_global_callback(cls, command_name, callback_function, command_id: Optional['snowflake.Snowflake'] = None) -> None:
        '''Register name of a command to a given callback function globally.'''
        cls.global_lookup[command_name] = callback_function
        if command_id is not None:
            cls.registered_commands[command_id] = callback_function

    @classmethod
    def register_guild_callback(cls,
                                command_name: str,
                                callback_function: 'Callable',
                                guild: 'Optional[Union[snowflake.Snowflake, ext_guild.Guild]]' = None,
                                command_id: 'Optional[snowflake.Snowflake]' = None
                                ) -> None:
        '''Register name of a command to a given callback function for guilds.

        Arguments:
            command_name (str): Name of the command to be registered. MUST match the name registered with Discord!
            callback_function (Callable): Callback function. Must be a function which accepts either (Interaction) or (Interaction, DiscordClient) as arguments.
            guild (Snowflake|Guild): Target guild to trigger against. Leave empty if all guilds will use this same named command.
            command_id (Snowflake): If known, register the command_id now. Saves an API call later.
        '''
        if isinstance(guild, ext_guild.Guild):
            guild = guild.id

        cls.guild_lookup[(command_name, guild)] = callback_function
        if command_id is not None:
            cls.registered_commands[command_id] = callback_function

    @classmethod
    def register_interaction_custom_id(cls, custom_id: str, callback_function: Callable, unlimited: bool = False):
        '''Register a callback for an interaction.

        Arguments:
            custom_id (str): The custom id to trigger on.
            callback_function (Callable): Function to call when interaction is triggered.
            unlimited (bool): Allows more than one interaction to be called. The user is expected to deregister this interaction later. If
                no cleanup occurs, the handler is cleaned after 15 minutes.
        '''
        # TODO: We can probably customize the duration of the cache at runtime?
        obj = CallbackData(custom_id, callback_function, unlimited)
        cls.registered_custom_ids[custom_id] = obj

    @classmethod
    def unregister_interaction_custom_id(cls, custom_id: str, not_exists_ok: bool = False):
        '''Attempt to remove a custom_id from the interaction cache. Warn if already removed.'''
        if custom_id in cls.registered_custom_ids:
            del cls.registered_custom_ids[custom_id]
        elif not not_exists_ok:
            raise KeyError(f'[{custom_id}] not in registered IDs!')

    @classmethod
    async def command_handler(cls, client, interaction: 'interactions.Interaction') -> None:  # noqa: C901
        '''Handle incoming commands and dispatch them to the correct type handler.'''
        if interaction.data is None:
            return

        if interaction.type == interactions.enumerations.INTERACTION_TYPES.MESSAGE_COMPONENT:
            await cls.handle_message_component(client, interaction)
        elif interaction.type == interactions.enumerations.INTERACTION_TYPES.APPLICATION_COMMAND:
            await cls.handle_application_command(client, interaction)

    @classmethod
    async def handle_application_command(cls, client, interaction: 'interactions.Interaction') -> None:  # noqa: C901
        '''Handle interactions against Messages and Users.'''
        assert interaction.data is not None
        key: Tuple[Any, Optional['snowflake.Snowflake']]

        cls._log.info(f'Attempt lookup of [{interaction.data.name}] with id [{interaction.data.id}].')

        if interaction.data.id in cls.registered_commands:
            cls._log.info(f'Found [{interaction.data.id}] in registered commands.')
            await cls._determine_args_and_call(cls.registered_commands[interaction.data.id], client, interaction)
            return

        # Lookup command in global
        try:
            results = await api.API.get_global_application_command(interaction.data.id)
            cls._log.info(f'API responded, attempt lookup inside global with [{results["name"]}].')
            if results['name'] in cls.global_lookup:
                cls.registered_commands[interaction.data.id] = cls.global_lookup[results['name']]
                cls._log.info(cls.registered_commands[interaction.data.id])
                cls._log.info(interaction.data.id)
                try:
                    await cls._determine_args_and_call(cls.registered_commands[interaction.data.id], client, interaction)
                except Exception as e:
                    raise e
                    raise RuntimeError(f'Callback function experienced an error {e}.')
                return
        except httpx.HTTPStatusError:
            pass

        # Lookup command in guild
        if (not hasattr(interaction, 'guild_id')) or (interaction.guild_id is None):
            raise LookupError(f'Unable to find interaction [{interaction.data.id}] in Global, and no guild_id is in object.')
        try:
            assert isinstance(interaction.guild_id, snowflake.Snowflake)
            results = await api.API.get_guild_application_command(interaction.guild_id, interaction.data.id)

            key = (results['name'], interaction.guild_id)
            cls._log.info('Attempt lookup inside guild.')
            if (key in cls.guild_lookup):
                cls.registered_commands[interaction.data.id] = cls.guild_lookup[key]
                try:
                    await cls._determine_args_and_call(cls.registered_commands[interaction.data.id], client, interaction)
                except Exception as e:
                    raise RuntimeError(f'Callback function experienced an error {e}.')
                return

            key = (results['name'], None)
            cls._log.info('Attempt lookup inside ALL guilds.')
            if (key in cls.guild_lookup):
                cls.registered_commands[interaction.data.id] = cls.guild_lookup[key]
                try:
                    await cls._determine_args_and_call(cls.registered_commands[interaction.data.id], client, interaction)
                except Exception as e:
                    raise RuntimeError(f'Callback function experienced an error {e}.')
                return
        except httpx.HTTPStatusError:
            pass

        raise LookupError(f'Unable to find interaction [{interaction.data.id}] in Global or Guild!')

    @classmethod
    async def _determine_args_and_call(cls, callable, client, interaction: 'interactions.Interaction') -> None:
        assert interaction.data is not None
        arg_len = len(inspect.signature(callable).parameters)
        if cls.__iscoroutinefunction_or_partial(callable):
            if arg_len == 1:
                await cls.registered_commands[interaction.data.id](interaction)
            elif arg_len == 2:
                await cls.registered_commands[interaction.data.id](interaction, client)
        else:
            if arg_len == 1:
                cls.registered_commands[interaction.data.id](interaction)
            elif arg_len == 2:
                cls.registered_commands[interaction.data.id](interaction, client)
            warnings.warn('While sync functions are supported as callbacks in CommandHandler, they are STRONGLY counter-recommended!', UserWarning)

    @classmethod
    async def handle_message_component(cls, client, interaction: 'interactions.Interaction') -> None:
        '''Handle slash-commands.'''
        assert interaction.data is not None
        cls._log.info(f'Saw id [{interaction.id}] with custom id [{interaction.data.custom_id}].')
        assert interaction.data is not None
        if interaction.data.custom_id not in cls.registered_custom_ids:
            raise RuntimeError('Got unexpected interaction')

        callback_data = cls.registered_custom_ids[interaction.data.custom_id]
        if callback_data.unlimited is False:
            del cls.registered_custom_ids[interaction.data.custom_id]

        if cls.__iscoroutinefunction_or_partial(callback_data.callback):
            await callback_data.callback(client, interaction)
        else:
            callback_data.callback(client, interaction)
            warnings.warn('While sync functions are supported as callbacks in CommandHandler, they are STRONGLY counter-recommended!', UserWarning)

    @classmethod
    def __iscoroutinefunction_or_partial(cls, object):
        while isinstance(object, functools.partial):
            object = object.func
        return inspect.iscoroutinefunction(object)



# import requests


# def auth_loop():
#     pass

# class auth_junk:
#     OAUTH_AUTH_RUL = 'https://discord.com/api/oauth2/authorize'
#     OAUTH_TOKEN_RUL = 'https://discord.com/api/oauth2/token'
#     OAUTH_REVOKE_RUL = 'https://discord.com/api/oauth2/token/revoke'

#     CLIENT_ID = '889065662641737791'

#     CLIENT_SECRET = 'a74X3B8Qaf3czzfIp54e8YYOlz_QziKu'

#     def __init__(self):
#         pass

#     @classmethod
#     def request_initial_auth(cls):
#         url = 'https://discord.com/api/oauth2/authorize?response_type=code&client_id=889065662641737791&scope=bot'
#         print(url)

#     @classmethod
#     def generate_bearer_token(cls, code):
#         REDIRECT_URI = 'http://localhost/redirect'

#         data = {
#             'client_id': cls.CLIENT_ID,
#             'client_secret': cls.CLIENT_SECRET,
#             'grant_type': 'authorization_code',
#             'code': code,
#             'redirect_uri': REDIRECT_URI
#         }
#         headers = {
#             'Content-Type': 'application/x-www-form-urlencoded'
#         }
#         r = httpx.post('%s/oauth2/token' % cls.BASE_URL, data=data, headers=headers)
#         r.raise_for_status()
#         return r.json()

#     @classmethod
#     def renew_bearer_token(cls):
#         pass



import enum


class MESSAGE_TYPE(enum.IntEnum):
    '''Types of messages.'''
    DEFAULT = 0
    RECIPIENT_ADD = 1
    RECIPIENT_REMOVE = 2
    CALL = 3
    CHANNEL_NAME_CHANGE = 4
    CHANNEL_ICON_CHANGE = 5
    CHANNEL_PINNED_MESSAGE = 6
    GUILD_MEMBER_JOIN = 7
    USER_PREMIUM_GUILD_SUBSCRIPTION = 8
    USER_PREMIUM_GUILD_SUBSCRIPTION_TIER_1 = 9
    USER_PREMIUM_GUILD_SUBSCRIPTION_TIER_2 = 10
    USER_PREMIUM_GUILD_SUBSCRIPTION_TIER_3 = 11
    CHANNEL_FOLLOW_ADD = 12
    GUILD_DISCOVERY_DISQUALIFIED = 14
    GUILD_DISCOVERY_REQUALIFIED = 15
    GUILD_DISCOVERY_GRACE_PERIOD_INITIAL_WARNING = 16
    GUILD_DISCOVERY_GRACE_PERIOD_FINAL_WARNING = 17
    THREAD_CREATED = 18
    REPLY = 19
    CHAT_INPUT_COMMAND = 20
    THREAD_STARTER_MESSAGE = 21
    GUILD_INVITE_REMINDER = 22
    CONTEXT_MENU_COMMAND = 23


class CHANNEL_TYPES(enum.IntEnum):
    '''Types of channels.'''
    GUILD_TEXT = 0  # a text channel within a server
    DM = 1  # a direct message between users
    GUILD_VOICE = 2  # a voice channel within a server
    GROUP_DM = 3  # a direct message between multiple users
    GUILD_CATEGORY = 4  # an organizational category that contains up to 50 channels
    GUILD_NEWS = 5  # a channel that users can follow and crosspost into their own server
    GUILD_STORE = 6  # a channel in which game developers can sell their game on Discord
    GUILD_NEWS_THREAD = 10  # a temporary sub-channel within a GUILD_NEWS channel
    GUILD_PUBLIC_THREAD = 11  # a temporary sub-channel within a GUILD_TEXT channel
    GUILD_PRIVATE_THREAD = 12  # a temporary sub-channel within a GUILD_TEXT channel that is only viewable by those invited and those with the MANAGE_THREADS permission
    GUILD_STAGE_VOICE = 13  # a voice channel for hosting events with an audience


class EMBED_TYPES(enum.Enum):
    '''Types of Embeds.'''
    rich = enum.auto()      # generic embed rendered from embed attributes
    image = enum.auto()     # image embed
    video = enum.auto()     # video embed
    gifv = enum.auto()      # animated gif image embed rendered as a video embed
    article = enum.auto()   # article embed
    link = enum.auto()      # link embed



import enum
import datetime
from typing import List

from . import snowflake, role
from .base_object import BaseDiscordObject


class User(BaseDiscordObject):
    '''Discord User.

    Attributes:
        id (Snowflake): Unique ID of User. Is a [Snowflake][dyscord.objects.snowflake.Snowflake].
        username (str): Global name of the User.
        discriminator (int): Random set of 4 numbers to discriminate user from others.
        avatar (str): TODO: Look this up.
        bot (bool): User is a bot or not.
    '''

    @enum.unique
    class UserFlag(enum.IntFlag):
        '''Flags on a User.'''
        NONE = 0
        DISCORD_EMPLOYEE = 1 << 0
        PARTNERED_SERVER_OWNER = 1 << 1
        HYPESQUAD_EVENTS = 1 << 2
        BUG_HUNTER_LEVEL_1 = 1 << 3
        HOUSE_BRAVERY = 1 << 6
        HOUSE_BRILLIANCE = 1 << 7
        HOUSE_BALANCE = 1 << 8
        EARLY_SUPPORTER = 1 << 9
        TEAM_USER = 1 << 10
        BUG_HUNTER_LEVEL_2 = 1 << 14
        VERIFIED_BOT = 1 << 16
        EARLY_VERIFIED_BOT_DEVELOPER = 1 << 17
        DISCORD_CERTIFIED_MODERATOR = 1 << 18

    @enum.unique
    class PremiumType(enum.IntFlag):
        '''Type of Premium subscription the User has.'''
        NONE = 0
        NITRO_CLASSIC = 1 << 0
        NITRO = 1 << 1

    id: snowflake.Snowflake = None  # type: ignore
    username: str = None  # type: ignore
    discriminator: str = None  # type: ignore
    avatar: str = None  # type: ignore
    bot: bool = None  # type: ignore
    system: bool = None  # type: ignore
    mfa_enabled: bool = None  # type: ignore
    banner: str = None  # type: ignore
    accent_color: int = None  # type: ignore
    locale: str = None  # type: ignore
    verified: bool = None  # type: ignore
    email: str = None  # type: ignore
    flags: int = None  # type: ignore
    premium_type: int = None  # type: ignore
    public_flags: int = None  # type: ignore

    def __str__(self):
        '''Return a discord compatible mention string.'''
        return self.mention_nickname

    def __repr__(self):
        '''Return a discord compatible mention string.'''
        return self.__str__()

    def __eq__(self, other):
        '''Determine if other User and self are the same User.'''
        if not isinstance(other, User):
            return NotImplemented
        return self.id == other.id

    # TODO: We could have a lot of handy "is_X" functions here to return if user is something form the USER_FLAG enumeration

    @property
    def mention(self):
        '''Generate a valid name mention for use in Discord.'''
        return f'<@{self.id}>'

    @property
    def mention_nickname(self):
        '''Generate a valid nickname mention for use in Discord.'''
        return f'<@!{self.id}>'

    def from_dict(self, data: dict) -> 'User':
        '''Parse a User from an API compliant dict.'''
        # Required fields
        self.id = snowflake.Snowflake(data['id'])
        self.username = data['username']
        self.discriminator = data['discriminator']
        self.avatar = data['avatar']

        # Optional fields
        if 'bot' in data:
            self.bot = data['bot']
        if 'system' in data:
            self.system = data['system']
        if 'verified' in data:
            self.verified = data['verified']
        if 'email' in data:
            self.email = data['email']
        if 'banner' in data:
            self.banner = data['banner']
        if 'accent_color' in data:
            self.accent_color = data['accent_color']
        if 'flags' in data:
            self.flags = data['flags']
        if 'premium_type' in data:
            self.premium_type = data['premium_type']
        if 'public_flags' in data:
            self.public_flags = data['public_flags']
        return self


class Member(User):
    '''User within a guild.'''

    nick: str = None  # type: ignore
    avatar: str = None  # type: ignore
    roles: List[role.Role] = None  # type: ignore
    joined_at: datetime.datetime = None  # type: ignore
    premium_since: datetime.datetime = None  # type: ignore
    deaf: bool = None  # type: ignore
    mute: bool = None  # type: ignore
    pending: bool = None  # type: ignore
    permissions: str = None  # type: ignore

    def __str__(self):
        '''Return string representation.'''
        fields = []

        if self.id is not None:
            fields.append(f'id={self.id}')

        if (self.username is not None) and (self.discriminator is not None):
            fields.append(f'username=\'{self.username}#{self.discriminator}\'')

        if self.nick is not None:
            fields.append(f'nick=\'{self.nick}\'')

        return f'Member({", ".join(fields)})'

    def from_dict(self, data: dict) -> 'Member':
        '''Parse a Member from an API compliant dict.'''
        if 'nick' in data:
            self.nick = data['nick']
        if 'avatar' in data:
            self.avatar = data['avatar']
        if 'joined_at' in data:
            self.joined_at = datetime.datetime.fromisoformat(data['joined_at'])
        if 'premium_since' in data and data['premium_since'] is not None:
            self.premium_since = datetime.datetime.fromisoformat(data['premium_since'])
        if 'deaf' in data:
            self.deaf = data['deaf']
        if 'mute' in data:
            self.mute = data['mute']
        if 'pending' in data:
            self.pending = data['pending']
        if 'permissions' in data:
            self.permissions = data['permissions']
        if 'user' in data:
            tmp_user = User().from_dict(data['user'])
            self.update_from_user(tmp_user)
        return self

    def update_from_user(self, user: User) -> 'Member':  # noqa: C901
        '''Update Member properties from associated user.'''
        if user.id is not None:
            self.id = user.id
        if user.username is not None:
            self.username = user.username
        if user.discriminator is not None:
            self.discriminator = user.discriminator
        if user.avatar is not None:
            self.avatar = user.avatar
        if user.bot is not None:
            self.bot = user.bot
        if user.system is not None:
            self.system = user.system
        if user.mfa_enabled is not None:
            self.mfa_enabled = user.mfa_enabled
        if user.banner is not None:
            self.banner = user.banner
        if user.accent_color is not None:
            self.accent_color = user.accent_color
        if user.locale is not None:
            self.locale = user.locale
        if user.verified is not None:
            self.verified = user.verified
        if user.email is not None:
            self.email = user.email
        if user.flags is not None:
            self.flags = user.flags
        if user.premium_type is not None:
            self.premium_type = user.premium_type
        if user.public_flags is not None:
            self.public_flags = user.public_flags
        return self



import asyncio
import functools
import inspect
import platform
import random
import sys
import time
import warnings

from collections import defaultdict
from pprint import pprint
from typing import Any, Callable, Optional, List

import nest_asyncio  # type: ignore
import orjson as json
import websockets

from . import api, INTENTS, DISCORD_EVENTS

from .. import utilities
from .. import objects
from .. import helper

from ..version import __version__

from .events import on_any, on_channel_create, on_channel_delete, on_channel_pins_update, on_channel_update, on_guild_ban_add, on_guild_ban_remove, on_guild_create,\
    on_guild_delete, on_guild_emojis_update, on_guild_integrations_update, on_guild_member_add, on_guild_member_remove, on_guild_member_update,\
    on_guild_role_create, on_guild_role_delete, on_guild_role_update, on_guild_stickers_update, on_guild_update, on_integration_create, on_integration_delete,\
    on_integration_update, on_invite_create, on_invite_delete, on_message_create, on_message_delete, on_message_delete_bulk, on_message_reaction_add,\
    on_message_reaction_remove, on_message_reaction_remove_all, on_message_reaction_remove_emoji, on_message_update, on_presence_update, on_ready,\
    on_stage_instance_create, on_stage_instance_delete, on_stage_instance_update, on_thread_create, on_thread_delete, on_thread_list_sync,\
    on_thread_member_update, on_thread_members_update, on_thread_update, on_typing_start, on_voice_state_update, on_webhooks_update, on_interaction_create


class DiscordClient:
    '''Client for interaction with Discord.'''

    _log = utilities.Log()
    _raw_callbacks: List[Callable] = list()
    _wrapper_registrations: dict = defaultdict(lambda: list())
    _wrapper_class_registrations: list = list()
    _version = __version__
    me: objects.User
    session_id: str
    token: str
    application_id: Optional[str]
    intent: int
    ready: bool

    def __init__(self, token: str, application_id: Optional[str] = None):
        '''Instantiate a DiscordClient.

        Args:
            token (str): Valid token to access discord. Only `Bot` token's currently supported.
            application_id (str): The application id. Can be left to None if client will not use Interactions.
        '''
        # Discord attributes
        self.__class__.token = token
        self.__class__.application_id = application_id
        self.__class__.intent = 0
        self.__class__.ready = False

        # Private attributes
        self._heartbeat_task = None
        self._last_heartbeat_ack = None
        self._listener_task = None
        self._sequence_number = None
        self._intents_defined = False

    def configure_intents(self,  # noqa: C901
                          guilds: bool = False,
                          guild_members: bool = False,
                          guild_bans: bool = False,
                          guild_emoji_and_stickers: bool = False,
                          guild_integrations: bool = False,
                          guild_webhooks: bool = False,
                          guild_invites: bool = False,
                          guild_voice_states: bool = False,
                          guild_presences: bool = False,
                          guild_messages: bool = False,
                          guild_message_reactions: bool = False,
                          guild_message_typeing: bool = False,
                          direct_messages: bool = False,
                          direct_message_reactions: bool = False,
                          direct_messages_typeing: bool = False,
                          ):
        '''Configure intents before connecting.

        For more details, read the [official docs](https://discord.com/developers/docs/topics/gateway#gateway-intents).

        Arguments:
            guilds (bool): Receive events related to guilds the client is a part of.
            guild_members (bool): Receive events related to guild members.
            guild_bans (bool): Receive events related to guild bans.
            guild_emoji_and_stickers (bool): Receive events related to guild emoji and stickers.
            guild_integrations (bool): Receive events related to guild Interactions.
            guild_webhooks (bool): Receive events related to guild Webhooks.
            guild_invites (bool): Receive events related to guild invites
            guild_voice_states (bool): Receive events related to guild voice states.
            guild_presences (bool): Receive events related to guild presences.
            guild_messages (bool): Receive events related to guild messages.
            guild_message_reactions (bool): Receive events related to guild reactions.
            guild_message_typeing (bool): Receive events related to guild members starting to type.
            direct_messages (bool): Receive events related to direct messages.
            direct_message_reactions (bool): Receive events related to direct message reactions.
            direct_messages_typeing (bool): Receive events related to users typing in direct messages.
        '''
        self.intent = 0
        if guilds:
            self.intent += INTENTS.GUILDS
        if guild_members:
            self.intent += INTENTS.GUILD_MEMBERS
        if guild_bans:
            self.intent += INTENTS.GUILD_BANS
        if guild_emoji_and_stickers:
            self.intent += INTENTS.GUILD_EMOJIS_AND_STICKERS
        if guild_integrations:
            self.intent += INTENTS.GUILD_INTEGRATIONS
        if guild_webhooks:
            self.intent += INTENTS.GUILD_WEBHOOKS
        if guild_invites:
            self.intent += INTENTS.GUILD_INVITES
        if guild_voice_states:
            self.intent += INTENTS.GUILD_VOICE_STATES
        if guild_presences:
            self.intent += INTENTS.GUILD_PRESENCES
        if guild_messages:
            self.intent += INTENTS.GUILD_MESSAGES
        if guild_message_reactions:
            self.intent += INTENTS.GUILD_MESSAGE_REACTIONS
        if guild_message_typeing:
            self.intent += INTENTS.GUILD_MESSAGE_TYPING
        if direct_messages:
            self.intent += INTENTS.DIRECT_MESSAGES
        if direct_message_reactions:
            self.intent += INTENTS.DIRECT_MESSAGE_REACTIONS
        if direct_messages_typeing:
            self.intent += INTENTS.DIRECT_MESSAGE_TYPING
        self._intents_defined = True

    def set_all_intents(self):
        '''Set all intents to True. For more information see the configure_intents() function.'''
        self.intent = 0
        for intent in INTENTS:
            self.intent += intent
        self._intents_defined = True

    def run(self, loop: asyncio.AbstractEventLoop = None):
        '''Start the async loop and run forever.

        Arguments:
            loop (asyncio.AbstractEventLoop): If desired, use a given asyncio compatible loop. One will be created if not given.
        '''
        self._log.info('Starting...')
        self._log.info(f'Platform: [{platform.platform()}]')
        self._log.info(f'Python Version: [{sys.version}]')
        self._log.info(f'Python Version Info: [{sys.version_info}]')
        self._log.info(f'Dyscord Version: [v{__version__}]')

        loop = loop if loop is not None else asyncio.get_event_loop()

        nest_asyncio.apply(loop)

        loop.create_task(self._run())

        loop.run_forever()

    async def _run(self, ):

        if self._intents_defined is False:
            warnings.warn('Started without defining intents. Client will likely get ZERO input. Consider calling the \'configure_intents\' function.', UserWarning)

        # BUG: This might cause Runtime errors, we need to wait and see. See https://github.com/erdewit/nest_asyncio/issues/22.
        # This will be left here for now, as it avoids weird issues and allows us to work if the user bypasses run().
        nest_asyncio.apply()

        loop = asyncio.get_event_loop()
        if not hasattr(loop, '_nest_patched'):
            raise RuntimeError('Cannot run this library without running \'nest_asyncio.apply()\' first.')

        # Start up the listener
        await self._connect()

        # Sleep forever
        while True:

            await asyncio.sleep(1)

            if self._listener_task is not None:

                if self._listener_task.done():

                    exception = self._listener_task.exception()
                    if exception:
                        try:
                            raise exception
                        except Exception:
                            self._log.exception('Caught exception')
                            raise

    async def _connect(self):
        '''TODO: Implement connection to discord's servers.'''
        api.API.TOKEN = self.token
        if type(self.application_id) is str:
            api.API.APPLICATION_ID = self.application_id

        gateway_uri = (await api.API.get_gateway_bot(self.token))['url']
        self._log.debug(f'Try to connect to {gateway_uri}')

        if self._listener_task is not None:
            self._listener_task.cancel()
            self._listener_task = None

        if self._heartbeat_task is not None:
            self._heartbeat_task.cancel()
            self._heartbeat_task = None
            self._last_heartbeat_ack = None

        self._listener_task = asyncio.create_task(self._web_socket_listener(gateway_uri))

        # Wait for heartbeat to start
        await asyncio.sleep(1)

        while self._last_heartbeat_ack is None:
            await asyncio.sleep(1)
            self._log.warning('Waiting on heartbeat...')

        self._log.debug('Heartbeat observed, begin to identify.')

        await self._identify()

    async def _web_socket_listener(self, uri):  # noqa

        def _handle_completed_tasks(task: asyncio.Task):
            exception = task.exception()
            if exception is None:
                return
            try:
                raise exception
            except Exception:
                self._log.exception('Exception from event dispatcher.')

        async with websockets.connect(uri) as websocket:

            self._gateway_ws = websocket

            while True:

                data = await websocket.recv()
                data = json.loads(data)

                for callback in self._raw_callbacks:
                    await callback(data)

                if 's' in data and data['s'] is not None:
                    self._sequence_number = data['s']
                    self._log.debug('Updated seq count to [{self._sequence_number}].')

                opcode = data['op']

                if opcode == 0:
                    self._log.debug('Received event, dispatch...')
                    task = asyncio.create_task(self._event_dispatcher(data))
                    task.add_done_callback(_handle_completed_tasks)

                elif opcode == 1:
                    self._log.debug('OPCODE: HEARTBEAT')
                    data = {'op': 1, 'd': self._sequence_number}
                    await self._gateway_ws.send(json.dumps(data))

                elif opcode == 7:
                    self._log.debug('OPCODE: RECONNECT')
                    await self._handle_op_7(data)

                elif opcode == 9:
                    self._log.debug('OPCODE: INVALID SESSION')
                    await self._handle_op_9(data)

                elif opcode == 10:
                    # Hello
                    self._log.debug('OPCODE: HELLO')
                    await self._handle_op_10(data)

                elif opcode == 11:
                    self._log.debug('Saw a heartbeat ACK')
                    self._last_heartbeat_ack = time.time()

                else:
                    self._log.error('Unknown opcode')
                    self._log.error(data)

    async def _heartbeat(self, interval):

        self._log.info('New heartbeat task started. Send new heartbeat NOW.')

        data = {'op': 1, 'd': self._sequence_number}
        await self._gateway_ws.send(json.dumps(data))

        while True:

            self._log.debug(f'Sleeping for {interval / 1000}s')

            await asyncio.sleep(interval / 1000)

            data = {'op': 1, 'd': self._sequence_number}
            self._log.debug(f'Sending heartbeat: {data}')

            await self._gateway_ws.send(json.dumps(data))

    async def _handle_op_7(self, data):

        await self._reconnect()

        self._log.critical('Opcode 7 handled.')

    async def _handle_op_9(self, data):

        if self._listener_task is not None:
            self._listener_task.cancel()
            self._listener_task = None

        if self._heartbeat_task is not None:
            self._heartbeat_task.cancel()
            self._heartbeat_task = None
            self._last_heartbeat_ack = None
            self._gateway_ws = None

        # Wait 1-5 seconds then try to connect.
        await asyncio.sleep(random.random() * 4 + 1)

        await self._reconnect()

        self._log.critical('Opcode 9 handled.')

    async def _handle_op_10(self, data):

        if self._heartbeat_task is not None:
            self._heartbeat_task.cancel()

        self._heartbeat_task = asyncio.create_task(self._heartbeat(data['d']['heartbeat_interval']))

        self._log.debug('Opcode 10 handled.')

    async def _reconnect(self):
        '''Send a reconnect message.'''
        gateway_uri = (await api.API.get_gateway_bot(self.token))['url']
        self._log.critical(f'Try to connect to {gateway_uri}')

        if self._listener_task is not None:
            self._listener_task.cancel()
            self._listener_task = None

        if self._heartbeat_task is not None:
            self._heartbeat_task.cancel()
            self._heartbeat_task = None
            self._last_heartbeat_ack = None
            self._gateway_ws = None

        self._listener_task = asyncio.create_task(self._web_socket_listener(gateway_uri))

        while self._gateway_ws is None:
            await asyncio.sleep(1)

        data = {
            'op': 6,
            'd': {
                'token': self.token,
                'session_id': self.session_id,
                'seq': self._sequence_number,
            }
        }
        self._log.info('Sending reconnect.')
        await self._gateway_ws.send(json.dumps(data))

    async def _identify(self):
        data = {
            'op': 2,
            'd': {
                'token': self.token,
                'intents': self.intent,
                "properties": {
                    "$os": "linux",
                    "$browser": "dyscord",
                    "$device": "dyscord"
                }
            }
        }
        self._log.info('Sending identify.')
        await self._gateway_ws.send(json.dumps(data))

    async def _event_dispatcher(self, data):  # noqa: C901

        event_type = data['t']
        self._log.info(f'Got a {event_type}')
        obj = None

        # pprint(data)

        if event_type == 'READY':
            obj = objects.Ready().from_dict(data['d'])
            self.__class__.session_id = obj.session_id
            self.__class__.ready = True
            self.__class__.me = obj.user
            self._log.info('Discord connection complete, we are ready!')
            self._log.info(f'We are now {self.me}')

        elif not self.__class__.ready:
            self._log.info(f'Got event of type [{event_type}] before we were ready!')
            return

        elif event_type == 'CHANNEL_CREATE':
            obj = objects.ChannelImporter().from_dict(data['d'])

        elif event_type == 'CHANNEL_DELETE':
            obj = objects.ChannelImporter().from_dict(data['d'])

        elif event_type == 'CHANNEL_PINS_UPDATE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'CHANNEL_UPDATE':
            obj = objects.ChannelImporter().from_dict(data['d'])

        elif event_type == 'GUILD_BAN_ADD':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'GUILD_BAN_REMOVE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'GUILD_CREATE':
            obj = objects.Guild().from_dict(data['d'])

        elif event_type == 'GUILD_DELETE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'GUILD_EMOJIS_UPDATE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'GUILD_INTEGRATIONS_UPDATE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'GUILD_MEMBER_ADD':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'GUILD_MEMBER_REMOVE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'GUILD_MEMBER_UPDATE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'GUILD_ROLE_CREATE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'GUILD_ROLE_DELETE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'GUILD_ROLE_UPDATE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'GUILD_STICKERS_UPDATE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'GUILD_UPDATE':
            obj = objects.Guild().from_dict(data['d'])

        elif event_type == 'INTEGRATION_CREATE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'INTEGRATION_DELETE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'INTEGRATION_UPDATE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'INVITE_CREATE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'INVITE_DELETE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'MESSAGE_CREATE':
            obj = objects.Message().from_dict(data['d'])

        elif event_type == 'MESSAGE_DELETE':
            obj = objects.Message().from_dict(data['d'])

        elif event_type == 'MESSAGE_DELETE_BULK':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'MESSAGE_REACTION_ADD':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'MESSAGE_REACTION_REMOVE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'MESSAGE_REACTION_REMOVE_ALL':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'MESSAGE_REACTION_REMOVE_EMOJI':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'MESSAGE_UPDATE':
            obj = objects.MessageUpdate().from_dict(data['d'])

        elif event_type == 'PRESENCE_UPDATE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'STAGE_INSTANCE_CREATE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'STAGE_INSTANCE_DELETE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'STAGE_INSTANCE_UPDATE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'THREAD_CREATE':
            obj = objects.ChannelImporter().from_dict(data['d'])

        elif event_type == 'THREAD_DELETE':
            obj = objects.ChannelImporter().from_dict(data['d'])

        elif event_type == 'THREAD_LIST_SYNC':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'THREAD_MEMBER_UPDATE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'THREAD_MEMBERS_UPDATE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'THREAD_UPDATE':
            obj = objects.ChannelImporter().from_dict(data['d'])

        elif event_type == 'TYPING_START':
            obj = objects.events.typing_start.TypingStart().from_dict(data['d'])

        elif event_type == 'VOICE_STATE_UPDATE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'WEBHOOKS_UPDATE':
            warnings.warn(f'Encountered unhandled event {event_type}')

        elif event_type == 'INTERACTION_CREATE':
            obj = objects.interactions.Interaction().from_dict(data['d'])
            self._log.info('Saw INTERACTION_CREATE event.')
            await helper.CommandHandler.command_handler(self, obj)

        else:
            # We have an unknown event on our hands, PANIC!!!
            self._log.critical(f'Encountered unknown event \'{data["t"]}\'!!!')
            pprint(data)
            return

        event_handler_name = f'on_{event_type.lower()}'

        # Call own event handlers first.
        if hasattr(self, event_handler_name):
            self_function = getattr(self, event_handler_name)
            arg_len = len(inspect.signature(self_function).parameters)
            arguments = (obj, data)
            await self_function(*arguments[:arg_len])

        # Call user wrapped classes, functions and cotoutines.
        # TODO: Should we invoke a create_task when able to avoid blocking calls?
        arguments = (obj, data, self)
        if obj is not None:
            for user_function in self.__class__._wrapper_registrations[event_type]:
                arg_len = len(inspect.signature(user_function).parameters)
                assert arg_len >= 0 and arg_len <= 3
                if asyncio.iscoroutinefunction(user_function):
                    await user_function(*arguments[:arg_len])
                else:
                    user_function(*arguments[:arg_len])

            for user_class in self.__class__._wrapper_class_registrations:
                if hasattr(user_class, event_handler_name):
                    user_function = getattr(user_class, event_handler_name)
                    arg_len = len(inspect.signature(user_function).parameters)
                    assert arg_len >= 0 and arg_len <= 4
                    if list(inspect.signature(user_function).parameters.items())[0][0] != 'cls':
                        warnings.warn('Wrapped class does not appear to be using class methods, unexpected behavior may result!', UserWarning)
                    if asyncio.iscoroutinefunction(user_function):
                        await user_function(user_class, *arguments[:arg_len])
                    else:
                        user_function(user_class, *arguments[:arg_len])

        # Handle the special case of the ANY event.
        for user_function in self.__class__._wrapper_registrations['ANY']:
            arg_len = len(inspect.signature(user_function).parameters)
            assert arg_len >= 0 and arg_len <= 3
            if asyncio.iscoroutinefunction(user_function):
                await user_function(*arguments[:arg_len])
            else:
                user_function(*arguments[:arg_len])

        for user_class in self.__class__._wrapper_class_registrations:
            if hasattr(user_class, 'on_any'):
                user_function = getattr(user_class, event_handler_name)
                arg_len = len(inspect.signature(user_function).parameters)
                assert arg_len >= 0 and arg_len <= 4
                if list(inspect.signature(user_function).parameters.items())[0][0] != 'cls':
                    warnings.warn('Wrapped class does not appear to be using class methods, unexpected behavior may result!', UserWarning)

                if asyncio.iscoroutinefunction(user_function):
                    await user_function(user_class, *arguments[:arg_len])
                else:
                    user_function(user_class, *arguments[:arg_len])

    # Register all out events
    on_any = on_any
    on_channel_create = on_channel_create
    on_channel_delete = on_channel_delete
    on_channel_pins_update = on_channel_pins_update
    on_channel_update = on_channel_update
    on_guild_ban_add = on_guild_ban_add
    on_guild_ban_remove = on_guild_ban_remove
    on_guild_create = on_guild_create
    on_guild_delete = on_guild_delete
    on_guild_emojis_update = on_guild_emojis_update
    on_guild_integrations_update = on_guild_integrations_update
    on_guild_member_add = on_guild_member_add
    on_guild_member_remove = on_guild_member_remove
    on_guild_member_update = on_guild_member_update
    on_guild_role_create = on_guild_role_create
    on_guild_role_delete = on_guild_role_delete
    on_guild_role_update = on_guild_role_update
    on_guild_stickers_update = on_guild_stickers_update
    on_guild_update = on_guild_update
    on_integration_create = on_integration_create
    on_integration_delete = on_integration_delete
    on_integration_update = on_integration_update
    on_invite_create = on_invite_create
    on_invite_delete = on_invite_delete
    on_message_create = on_message_create
    on_message_delete = on_message_delete
    on_message_delete_bulk = on_message_delete_bulk
    on_message_reaction_add = on_message_reaction_add
    on_message_reaction_remove = on_message_reaction_remove
    on_message_reaction_remove_all = on_message_reaction_remove_all
    on_message_reaction_remove_emoji = on_message_reaction_remove_emoji
    on_message_update = on_message_update
    on_presence_update = on_presence_update
    on_ready = on_ready
    on_stage_instance_create = on_stage_instance_create
    on_stage_instance_delete = on_stage_instance_delete
    on_stage_instance_update = on_stage_instance_update
    on_thread_create = on_thread_create
    on_thread_delete = on_thread_delete
    on_thread_list_sync = on_thread_list_sync
    on_thread_member_update = on_thread_member_update
    on_thread_members_update = on_thread_members_update
    on_thread_update = on_thread_update
    on_typing_start = on_typing_start
    on_voice_state_update = on_voice_state_update
    on_webhooks_update = on_webhooks_update
    on_interaction_create = on_interaction_create

    @classmethod
    def decorate_handler(cls, event: str):
        '''Register a given function to a given event string.

        This function should be used as a decorator around a function to map that function to a given event. The decorator takes one argument, a string which maps to the type of event we should map
        to. See the definition of DISCORD_EVENTS for names to map against.

        The decorated mapped should by an `async` function, although synchronous functions are allowed (and highly discouraged).

        The arguments of the decorated decide which python objects are given to the function when the given event is called. The mapping is as follows:

        - 1 Argument -> (discord_object)
        - 2 Argument -> (discord_object, raw_dict)
        - 3 Argument -> (discord_object, raw_dict, client)

        The `discord_object` will be a python representation of the event object.

        The `raw_dict` is a raw dictionary the API emitted.

        The `client` is the DiscordClient instance.
        '''
        if not hasattr(DISCORD_EVENTS, event) and event != 'ANY':
            raise ValueError(f'Attempted to bind to unknown event \'{event}\', must be exact match for existing {DISCORD_EVENTS} entry.')

        def func_wrapper(func):
            arg_len = len(inspect.signature(func).parameters)
            assert arg_len >= 0 and arg_len <= 3
            if asyncio.iscoroutinefunction(func):
                @functools.wraps(func)
                async def wrapped_func(*args, **kwargs):  # type: ignore
                    await func(*args, **kwargs)

                cls._wrapper_registrations[event].append(wrapped_func)
                return wrapped_func
            else:
                @functools.wraps(func)
                def wrapped_func(*args, **kwargs):  # type: ignore
                    func(*args, **kwargs)

                cls._wrapper_registrations[event].append(wrapped_func)
                return wrapped_func

        return func_wrapper

    @classmethod
    def decorate_class(cls, target_class):
        '''Register a given class and attempt to call any valid on_<event> functions.

        By convention functions of the class should be async.
        '''

        @functools.wraps(target_class)
        async def class_wrapper(cls, *args, **kwargs):
            pass

        cls._wrapper_class_registrations.append(class_wrapper)

        return class_wrapper

    @classmethod
    def _register_raw_callback(cls, callback: Callable[[dict], Any]):
        '''Register a raw callback that will receive pure dicts from the API.'''
        cls._raw_callbacks.append(callback)



'''Borg design module.'''
from typing import Dict


class Borg:
    '''Borg design pattern base class.'''

    _shared_state: Dict[object, object] = {}

    def __init__(self):
        '''To be called by child-classes.'''
        self.__dict__: Dict[object, object] = self._shared_state




import enum


class INTENTS(enum.IntFlag):
    '''Intents of the user.'''

    GUILDS = (1 << 0)
    # GUILD_CREATE = (1 << 0)
    # GUILD_UPDATE = (1 << 0)
    # GUILD_DELETE = (1 << 0)
    # GUILD_ROLE_CREATE = (1 << 0)
    # GUILD_ROLE_UPDATE = (1 << 0)
    # GUILD_ROLE_DELETE = (1 << 0)
    # CHANNEL_CREATE = (1 << 0)
    # CHANNEL_UPDATE = (1 << 0)
    # CHANNEL_DELETE = (1 << 0)
    # CHANNEL_PINS_UPDATE = (1 << 0)
    # THREAD_CREATE = (1 << 0)
    # THREAD_UPDATE = (1 << 0)
    # THREAD_DELETE = (1 << 0)
    # THREAD_LIST_SYNC = (1 << 0)
    # THREAD_MEMBER_UPDATE = (1 << 0)
    # THREAD_MEMBERS_UPDATE = (1 << 0)
    # STAGE_INSTANCE_CREATE = (1 << 0)
    # STAGE_INSTANCE_UPDATE = (1 << 0)
    # STAGE_INSTANCE_DELETE = (1 << 0)

    GUILD_MEMBERS = (1 << 1)
    # GUILD_MEMBER_ADD = (1 << 1)
    # GUILD_MEMBER_UPDATE = (1 << 1)
    # GUILD_MEMBER_REMOVE = (1 << 1)
    # THREAD_MEMBERS_UPDATE = (1 << 1)

    GUILD_BANS = (1 << 2)
    # GUILD_BAN_ADD = (1 << 2)
    # GUILD_BAN_REMOVE = (1 << 2)

    GUILD_EMOJIS_AND_STICKERS = (1 << 3)
    # GUILD_EMOJIS_UPDATE = (1 << 3)
    # GUILD_STICKERS_UPDATE = (1 << 3)

    GUILD_INTEGRATIONS = (1 << 4)
    # GUILD_INTEGRATIONS_UPDATE = (1 << 4)
    # INTEGRATION_CREATE = (1 << 4)
    # INTEGRATION_UPDATE = (1 << 4)
    # INTEGRATION_DELETE = (1 << 4)

    GUILD_WEBHOOKS = (1 << 5)
    # WEBHOOKS_UPDATE = (1 << 5)

    GUILD_INVITES = (1 << 6)
    # INVITE_CREATE = (1 << 6)
    # INVITE_DELETE = (1 << 6)

    GUILD_VOICE_STATES = (1 << 7)
    # VOICE_STATE_UPDATE = (1 << 7)

    GUILD_PRESENCES = (1 << 8)
    # PRESENCE_UPDATE = (1 << 8)

    GUILD_MESSAGES = (1 << 9)
    # MESSAGE_CREATE = (1 << 9)
    # MESSAGE_UPDATE = (1 << 9)
    # MESSAGE_DELETE = (1 << 9)
    # MESSAGE_DELETE_BULK = (1 << 9)

    GUILD_MESSAGE_REACTIONS = (1 << 10)
    # MESSAGE_REACTION_ADD = (1 << 10)
    # MESSAGE_REACTION_REMOVE = (1 << 10)
    # MESSAGE_REACTION_REMOVE_ALL = (1 << 10)
    # MESSAGE_REACTION_REMOVE_EMOJI = (1 << 10)

    GUILD_MESSAGE_TYPING = (1 << 11)
    # TYPING_START = (1 << 11)

    DIRECT_MESSAGES = (1 << 12)
    # MESSAGE_CREATE = (1 << 12)
    # MESSAGE_UPDATE = (1 << 12)
    # MESSAGE_DELETE = (1 << 12)
    # CHANNEL_PINS_UPDATE = (1 << 12)

    DIRECT_MESSAGE_REACTIONS = (1 << 13)
    # MESSAGE_REACTION_ADD = (1 << 13)
    # MESSAGE_REACTION_REMOVE = (1 << 13)
    # MESSAGE_REACTION_REMOVE_ALL = (1 << 13)
    # MESSAGE_REACTION_REMOVE_EMOJI = (1 << 13)

    DIRECT_MESSAGE_TYPING = (1 << 14)
    # TYPING_START = (1 << 14)


class DISCORD_EVENTS(enum.Enum):
    '''Enumeration of various discord events.

    Attributes:
        CHANNEL_CREATE (enum):
        CHANNEL_DELETE (enum):
        CHANNEL_PINS_UPDATE (enum):
        CHANNEL_UPDATE (enum):
        GUILD_BAN_ADD (enum):
        GUILD_BAN_REMOVE (enum):
        GUILD_CREATE (enum):
        GUILD_DELETE (enum):
        GUILD_EMOJIS_UPDATE (enum):
        GUILD_INTEGRATIONS_UPDATE (enum):
        GUILD_MEMBER_ADD (enum):
        GUILD_MEMBER_REMOVE (enum):
        GUILD_MEMBER_UPDATE (enum):
        GUILD_ROLE_CREATE (enum):
        GUILD_ROLE_DELETE (enum):
        GUILD_ROLE_UPDATE (enum):
        GUILD_STICKERS_UPDATE (enum):
        GUILD_UPDATE (enum):
        INTEGRATION_CREATE (enum):
        INTEGRATION_DELETE (enum):
        INTEGRATION_UPDATE (enum):
        INVITE_CREATE (enum):
        INVITE_DELETE (enum):
        MESSAGE_CREATE (enum):
        MESSAGE_DELETE (enum):
        MESSAGE_DELETE_BULK (enum):
        MESSAGE_REACTION_ADD (enum):
        MESSAGE_REACTION_REMOVE (enum):
        MESSAGE_REACTION_REMOVE_ALL (enum):
        MESSAGE_REACTION_REMOVE_EMOJI (enum):
        MESSAGE_UPDATE (enum):
        PRESENCE_UPDATE (enum):
        READY (enum):
        STAGE_INSTANCE_CREATE (enum):
        STAGE_INSTANCE_DELETE (enum):
        STAGE_INSTANCE_UPDATE (enum):
        THREAD_CREATE (enum):
        THREAD_DELETE (enum):
        THREAD_LIST_SYNC (enum):
        THREAD_MEMBER_UPDATE (enum):
        THREAD_MEMBERS_UPDATE (enum):
        THREAD_UPDATE (enum):
        TYPING_START (enum):
        VOICE_STATE_UPDATE (enum):
        WEBHOOKS_UPDATE (enum):
        INTERACTION_CREATE (enum):
    '''
    CHANNEL_CREATE = enum.auto()
    CHANNEL_DELETE = enum.auto()
    CHANNEL_PINS_UPDATE = enum.auto()
    CHANNEL_UPDATE = enum.auto()
    GUILD_BAN_ADD = enum.auto()
    GUILD_BAN_REMOVE = enum.auto()
    GUILD_CREATE = enum.auto()
    GUILD_DELETE = enum.auto()
    GUILD_EMOJIS_UPDATE = enum.auto()
    GUILD_INTEGRATIONS_UPDATE = enum.auto()
    GUILD_MEMBER_ADD = enum.auto()
    GUILD_MEMBER_REMOVE = enum.auto()
    GUILD_MEMBER_UPDATE = enum.auto()
    GUILD_ROLE_CREATE = enum.auto()
    GUILD_ROLE_DELETE = enum.auto()
    GUILD_ROLE_UPDATE = enum.auto()
    GUILD_STICKERS_UPDATE = enum.auto()
    GUILD_UPDATE = enum.auto()
    INTEGRATION_CREATE = enum.auto()
    INTEGRATION_DELETE = enum.auto()
    INTEGRATION_UPDATE = enum.auto()
    INVITE_CREATE = enum.auto()
    INVITE_DELETE = enum.auto()
    MESSAGE_CREATE = enum.auto()
    MESSAGE_DELETE = enum.auto()
    MESSAGE_DELETE_BULK = enum.auto()
    MESSAGE_REACTION_ADD = enum.auto()
    MESSAGE_REACTION_REMOVE = enum.auto()
    MESSAGE_REACTION_REMOVE_ALL = enum.auto()
    MESSAGE_REACTION_REMOVE_EMOJI = enum.auto()
    MESSAGE_UPDATE = enum.auto()
    PRESENCE_UPDATE = enum.auto()
    READY = enum.auto()
    STAGE_INSTANCE_CREATE = enum.auto()
    STAGE_INSTANCE_DELETE = enum.auto()
    STAGE_INSTANCE_UPDATE = enum.auto()
    THREAD_CREATE = enum.auto()
    THREAD_DELETE = enum.auto()
    THREAD_LIST_SYNC = enum.auto()
    THREAD_MEMBER_UPDATE = enum.auto()
    THREAD_MEMBERS_UPDATE = enum.auto()
    THREAD_UPDATE = enum.auto()
    TYPING_START = enum.auto()
    VOICE_STATE_UPDATE = enum.auto()
    WEBHOOKS_UPDATE = enum.auto()
    INTERACTION_CREATE = enum.auto()



'''Dyscord is a library to enable a reasonably 1:1 binding between python and the official API.'''

from . import client, helper, objects, utilities, command
from .version import __version__
from .client.discord_client import DiscordClient

__all__ = [
    '__version__',
    'client',
    'DiscordClient',
    'helper',
    'objects',
    'utilities',
    'command',
]



import logging


class Log:
    '''Generalized logging class for easy import.'''

    log = logging.getLogger('dyscord')
    log.setLevel(logging.INFO)
    _ch = logging.StreamHandler()
    _formatter = logging.Formatter('{asctime} - {levelname} - {filename}:{lineno} - {funcName} - {message}', style='{')
    _ch.setFormatter(_formatter)
    log.addHandler(_ch)

    def __getattr__(self, name):
        '''Map in the logging class attributes.'''
        return getattr(self.log, name)



'''Components that can be attached to messages.'''

import abc
import uuid
from typing import Callable, Dict, List, Union, Optional

from . import command, enumerations
from ..base_object import BaseDiscordObject
from .. import emoji
from ...helper import command_handler


class Component(BaseDiscordObject, abc.ABC):
    '''Base class for Components.'''
    type: enumerations.COMPONENT_TYPES


class ComponentAdder(abc.ABC):
    '''Allow other objects to start adding components to themselves with a common set of helper functions.

    Caution: This is an abstract class, and is not intended for direct instantiation.
    '''

    def add_components(self) -> 'ActionRow':
        '''Start adding components by starting an ACTION_ROW.'''
        if not hasattr(self, 'components') or self.components is None:
            self.components: Optional[List['Component']] = list()
        assert type(self.components) is list
        new_action_row = ActionRow()
        self.components.append(new_action_row)

        return new_action_row


class ActionRow(Component):
    '''Intermediate Component that holds Button or SelectMenu objects.'''
    type = enumerations.COMPONENT_TYPES.ACTION_ROW
    components: List[Union['Button', 'SelectMenu']]

    BUTTON_STYLES = enumerations.BUTTON_STYLES

    def add_button(self,
                   style: enumerations.BUTTON_STYLES,
                   custom_id: Optional[str] = None,
                   label: Optional[str] = None,
                   emoji: Optional[emoji.Emoji] = None,
                   url: Optional[str] = None,
                   disabled: Optional[bool] = None,
                   callback: Callable = None,
                   unlimited_callbacks: bool = False,
                   ) -> 'Button':
        '''Append a Button item to the ActionRow, and return it.

        Arguments:
            style (BUTTON_STYLES): Style of button to be used.
            custom_id (str): Unique string to be returned by the API when item is selected. Default is UUID4.
            label (str): Actual text user will see.
            emoji (Emoji): Emoji rather than label to display.
            url: (str): URL for BUTTON_STYLES.LINK buttons.
            disabled (bool): Item will display, but is not interactable.
            callback (Callable): Callback function when user selects an item. Should be
            unlimited_callbacks (bool): Request callbacks even if custom_id has already been called.

        Returns:
            Button
        '''
        if not hasattr(self, 'components'):
            self.components = list()
        new_button = Button()
        new_button.style = style
        if label is not None:
            new_button.label = label
        if emoji is not None:
            new_button.emoji = emoji
        if custom_id is not None:
            new_button.custom_id = custom_id
        elif custom_id is None and style != enumerations.BUTTON_STYLES.LINK:
            new_button.custom_id = str(uuid.uuid4())
        if url is not None:
            new_button.url = url
        if disabled is not None:
            new_button.disabled = disabled

        self.components.append(new_button)

        if callback is not None:
            command_handler.CommandHandler.register_interaction_custom_id(
                new_button.custom_id,
                callback,
                unlimited_callbacks,
            )

        return new_button

    def add_select_menu(self,
                        custom_id: Optional[str] = None,
                        placeholder: Optional[str] = None,
                        min_values: Optional[int] = None,
                        max_values: Optional[int] = None,
                        disabled: Optional[bool] = None,
                        callback: Callable = None,
                        unlimited_callbacks: bool = False,
                        ) -> 'SelectMenu':
        '''Append a SelectMenu item to the ActionRow, and return it.

        Arguments:
            custom_id (str): Unique string to be returned by the API when item is selected. Default is UUID4.
            placeholder (str): Text to appear in the box before user selection begins.
            min_values (int): Min number of items user must select. Default is 1.
            max_values (int): Max number of items user can select. Default is 1.
            disabled (bool): Item will display, but is not interactable.
            callback (Callable): Callback function when user selects an item. Should be
            unlimited_callbacks (bool): Request callbacks even if custom_id has already been called.

        Returns:
            SelectMenu
        '''
        if not hasattr(self, 'components'):
            self.components = list()
        new_select_menu = SelectMenu()
        if custom_id is not None:
            new_select_menu.custom_id = custom_id
        else:
            new_select_menu.custom_id = str(uuid.uuid4())
        if placeholder is not None:
            new_select_menu.placeholder = placeholder
        if min_values is not None:
            new_select_menu.min_values = min_values
        if max_values is not None:
            new_select_menu.max_values = max_values
        if disabled is not None:
            new_select_menu.disabled = disabled

        self.components.append(new_select_menu)

        if callback is not None:
            command_handler.CommandHandler.register_interaction_custom_id(
                new_select_menu.custom_id,
                callback,
                unlimited_callbacks,
            )

        return new_select_menu

    def to_dict(self) -> dict:
        '''Convert object to dictionary suitable for API or other generic useage.'''
        new_dict: Dict[str, object] = dict()
        new_dict['type'] = self.type.value
        new_dict['components'] = list()
        assert type(new_dict['components']) is list
        for component in self.components:
            new_dict['components'].append(component.to_dict())

        return new_dict

    def validate(self):
        '''Validate object is prepared for dispatch to discord.'''
        if hasattr(self, 'components'):
            assert type(self.components) is list,\
                f'Got invalid type {type(self.components)} of ActionRow.components, must be list.'

            assert len(self.components) <= 5,\
                f'Found {len(self.components)} components, discord allows a max of 5.'

            for component in self.components:
                assert type(component) in [Button, SelectMenu],\
                    f'Got invalid type {type(component)} in ActionRow.components, must be Union[Button, SelectMenu].'
                component.validate()


class Button(Component):
    '''Selectable button or link for user interaction.'''

    BUTTON_STYLES = enumerations.BUTTON_STYLES

    type = enumerations.COMPONENT_TYPES.BUTTON
    custom_id: str
    disabled: bool
    style: enumerations.BUTTON_STYLES
    label: str
    emoji: emoji.Emoji
    url: str

    def to_dict(self) -> dict:
        '''Convert object to dictionary suitable for API or other generic useage.'''
        new_dict: Dict[str, object] = dict()
        new_dict['type'] = self.type.value
        if hasattr(self, 'custom_id'):
            new_dict['custom_id'] = self.custom_id
        if hasattr(self, 'disabled'):
            new_dict['disabled'] = self.disabled
        if hasattr(self, 'style'):
            new_dict['style'] = self.style
        if hasattr(self, 'label'):
            new_dict['label'] = self.label
        if hasattr(self, 'emoji'):
            new_dict['emoji'] = self.emoji
        if hasattr(self, 'url'):
            new_dict['url'] = self.url
        return new_dict

    def validate(self):
        '''Validate object is prepared for dispatch to discord.'''
        assert self.type == enumerations.COMPONENT_TYPES.BUTTON,\
            f'Got invalid value {self.type} of Button.type, must be {enumerations.COMPONENT_TYPES.BUTTON}.'
        assert type(self.type) is enumerations.COMPONENT_TYPES,\
            f'Got invalid type {type(self.type)} of Button.type, must be {enumerations.COMPONENT_TYPES}.'

        assert hasattr(self, 'style'),\
            'A style must be set for Button.'
        assert type(self.style) is enumerations.BUTTON_STYLES,\
            f'Got invalid type {type(self.style)} of Button.style, must be {enumerations.BUTTON_STYLES}.'

        if self.style != enumerations.BUTTON_STYLES.LINK:
            assert hasattr(self, 'label'),\
                f'Buttons in style {self.style} must have a label.'
            assert type(self.label) is str,\
                f'Got invalid type {type(self.label)} of Button.label, must be str.'
            assert len(self.label) <= 80,\
                f'Got Button.label length of {len(self.label)}, max is 80.'
            assert len(self.label) >= 1,\
                f'Got Button.label length of {len(self.label)}, min is 1.'
            assert not hasattr(self, 'url'),\
                f'Cannot have url in a {self.style}.'
            assert hasattr(self, 'custom_id'),\
                'Button must have a custom_id.'
            assert type(self.custom_id) is str,\
                f'Got invalid type {type(self.custom_id)} of Button.custom_id, must be str.'
        else:
            assert hasattr(self, 'url'),\
                'Button must have URL when style is BUTTON_STYLES.LINK.'


class SelectMenu(Component):
    '''A drop down menu for user interaction.'''

    type = enumerations.COMPONENT_TYPES.SELECT_MENU
    custom_id: str
    disabled: bool
    options: List['command.CommandOptions']
    placeholder: str
    min_values: int
    max_values: int

    add_option_typed = command.Command.add_option_typed

    def to_dict(self) -> dict:
        '''Convert object to dictionary suitable for API or other generic useage.'''
        new_dict: Dict[str, object] = dict()
        new_dict['type'] = self.type.value
        if hasattr(self, 'custom_id'):
            new_dict['custom_id'] = self.custom_id
        if hasattr(self, 'disabled'):
            new_dict['disabled'] = self.disabled
        if hasattr(self, 'options'):
            new_dict['options'] = list()
            assert type(new_dict['options']) is list
            for option in self.options:
                new_dict['options'].append(option.to_dict())
        if hasattr(self, 'placeholder'):
            new_dict['placeholder'] = self.placeholder
        if hasattr(self, 'min_values'):
            new_dict['min_values'] = self.min_values
        if hasattr(self, 'max_values'):
            new_dict['max_values'] = self.max_values
        return new_dict

    def validate(self):
        '''Validate object is prepared for dispatch to discord.'''
        pass



from .emoji import Emoji
from .guild import Guild
from .message import Message, MessageUpdate
from .permissions import Permissions
from .ready import Ready
from .snowflake import Snowflake
from .user import User, Member
from .role import Role
from .embed import Embed, EmbedAdder

from .channel import Channel, ChannelImporter, CategoryChannel, NewsChannel, TextChannel, StoreChannel, VoiceChannel

from . import interactions
from . import events

__all__ = [
    'CategoryChannel',
    'Channel',
    'ChannelImporter',
    'Embed',
    'EmbedAdder',
    'Emoji',
    'events',
    'Guild',
    'interactions',
    'Member',
    'Message',
    'MessageUpdate',
    'NewsChannel',
    'Permissions',
    'Ready',
    'Role',
    'Snowflake',
    'StoreChannel',
    'TextChannel',
    'User',
    'VoiceChannel',
]



import copy
from typing import List

from .base_object import BaseDiscordObject
from . import user as ext_user, snowflake, guild


class Ready(BaseDiscordObject):
    '''Event sent from servers when Discord is ready and connected.'''

    # application: Dict[str, Union[int, str]]  # TODO: Application objects needs to exist.
    geo_ordered_rtc_regions: List[str]
    guild_join_requests: list  # TODO: See examples of this and process.
    guilds: List[guild.Guild]
    presences: list  # TODO: Get examples of this.
    private_channels: list  # TODO: Get examples of this.
    relationships: list  # TODO: Get examples of this.
    session_id: str  # Why is this not a snowflake? Dammit!
    shard_id: int  # TODO: Get examples of this.
    shard_total: int  # TODO: Get examples of this.
    user: 'ext_user.User'
    user_settings: dict  # TODO: Get example of this.
    version: int  # Version of the API being used.

    def __init__(self):
        '''We don't have an ID for these objects, so do not call the super.'''
        pass

    def __hash__(self):
        '''Hash.'''
        return hash(self.session_id)

    def __eq__(self, other):
        '''Determine if other object came form the same application.'''
        return self.session_id == other.session_id

    def from_dict(self, data: dict) -> 'Ready':
        '''Parse a Ready from an API compliant dict.'''
        self.geo_ordered_rtc_regions = copy.deepcopy(data['geo_ordered_rtc_regions'])
        self.guilds = list()
        for partial_guild_dict in data['guilds']:
            new_guild = guild.Guild()
            new_guild.id = snowflake.Snowflake(partial_guild_dict['id'])
            self.guilds.append(new_guild)
        self.session_id = data['session_id']
        self.user = ext_user.User().from_dict(data['user'])
        self.version = data['v']

        return self



'''Cache values for discord to use. NOT CURRENTLY USED.'''

from .. import objects
from .borg import Borg


class Cache(Borg):
    '''Generic cache of objects we have been told about from the API.'''

    def __init__(self):
        '''Create a access to the cache, follows the Borg design pattern.'''
        super().__init__()

        try:
            getattr(self, 'first')
        except AttributeError:
            self.guilds = set()
            self.channels = set()
            self.users = set()
            self.first = False

    def get(self, identifier: 'objects.Snowflake'):
        '''Return a given object with an given identifier if we know about it.'''
        for x in self.guilds:
            if identifier == x.id:
                return x
        for x in self.channels:
            if identifier == x.id:
                return x
        for x in self.users:
            if identifier == x.id:
                return x
        raise LookupError(f'Identifier {identifier} not in cache.')

    def add(self, object):
        '''Given some generic object, insert it into the cache.'''
        # Convert to dict
        # if type(object) is objects.Guild:
        #     self.guilds.add(object)
        # elif isinstance(object, objects.Channel):
        #     self.channels.add(object)
        # elif type(object) is objects.User:
        #     self.users.add(object)

    def clear(self):
        '''Remove all current elements within the cache.'''
        self.guilds = set()
        self.channels = set()
        self.users = set()



import enum
from typing import Optional, Union


class Permissions:
    '''Permissions for users and roles.'''

    class PermissionFlags(enum.IntFlag):
        '''Various flags that can be set.'''

        CREATE_INSTANT_INVITE = (1 << 0)        # Allows creation of instant invites	T, V, S
        KICK_MEMBERS = (1 << 1)                 # Allows kicking members
        BAN_MEMBERS = (1 << 2)                  # Allows banning members
        ADMINISTRATOR = (1 << 3)                # Allows all permissions and bypasses channel permission overwrites
        MANAGE_CHANNELS = (1 << 4)              # Allows management and editing of channels	T, V, S
        MANAGE_GUILD = (1 << 5)                 # Allows management and editing of the guild
        ADD_REACTIONS = (1 << 6)                # Allows for the addition of reactions to messages	T
        VIEW_AUDIT_LOG = (1 << 7)               # Allows for viewing of audit logs
        PRIORITY_SPEAKER = (1 << 8)             # Allows for using priority speaker in a voice channel	V
        STREAM = (1 << 9)                       # Allows the user to go live	V
        VIEW_CHANNEL = (1 << 10)                # Allows guild members to view a channel, which includes reading messages in text channels	T, V, S
        SEND_MESSAGES = (1 << 11)               # Allows for sending messages in a channel (does not allow sending messages in threads)	T
        SEND_TTS_MESSAGES = (1 << 12)           # Allows for sending of /tts messages	T
        MANAGE_MESSAGES = (1 << 13)             # Allows for deletion of other users messages	T
        EMBED_LINKS = (1 << 14)                 # Links sent by users with this permission will be auto-embedded	T
        ATTACH_FILES = (1 << 15)                # Allows for uploading images and files	T
        READ_MESSAGE_HISTORY = (1 << 16)        # Allows for reading of message history	T
        MENTION_EVERYONE = (1 << 17)            # Allows for using the @everyone tag to notify all users in a channel, and the @here tag to notify all online users in a channel	T
        USE_EXTERNAL_EMOJIS = (1 << 18)         # Allows the usage of custom emojis from other servers	T
        VIEW_GUILD_INSIGHTS = (1 << 19)         # Allows for viewing guild insights
        CONNECT = (1 << 20)                     # Allows for joining of a voice channel	V, S
        SPEAK = (1 << 21)                       # Allows for speaking in a voice channel	V
        MUTE_MEMBERS = (1 << 22)                # Allows for muting members in a voice channel	V, S
        DEAFEN_MEMBERS = (1 << 23)              # Allows for deafening of members in a voice channel	V, S
        MOVE_MEMBERS = (1 << 24)                # Allows for moving of members between voice channels	V, S
        USE_VAD = (1 << 25)                     # Allows for using voice-activity-detection in a voice channel	V
        CHANGE_NICKNAME = (1 << 26)             # Allows for modification of own nickname
        MANAGE_NICKNAMES = (1 << 27)            # Allows for modification of other users nicknames
        MANAGE_ROLES = (1 << 28)                # Allows management and editing of roles	T, V, S
        MANAGE_WEBHOOKS = (1 << 29)             # Allows management and editing of webhooks	T
        MANAGE_EMOJIS_AND_STICKERS = (1 << 30)  # Allows management and editing of emojis and stickers
        USE_APPLICATION_COMMANDS = (1 << 31)    # Allows members to use application commands, including slash commands and context menu commands.	T
        REQUEST_TO_SPEAK = (1 << 32)            # Allows for requesting to speak in stage channels. (This permission is under active development and may be changed or removed.)	S
        MANAGE_THREADS = (1 << 34)              # Allows for deleting and archiving threads, and viewing all private threads	T
        CREATE_PUBLIC_THREADS = (1 << 35)       # Allows for creating threads	T
        CREATE_PRIVATE_THREADS = (1 << 36)      # Allows for creating private threads	T
        USE_EXTERNAL_STICKERS = (1 << 37)       # Allows the usage of custom stickers from other servers	T
        SEND_MESSAGES_IN_THREADS = (1 << 38)    # Allows for sending messages in threads	T
        START_EMBEDDED_ACTIVITIES = (1 << 39)   # Allows for launching activities (applications with the EMBEDDED flag) in a voice channel	V

    def __init__(self, data: Optional[Union[int, str, 'Permissions']] = None):
        '''Init.'''
        self.permissions: set

        if type(data) in [str, int]:
            self.parse_str(str(data))
        elif type(data) is Permissions:
            self.permissions = data.permissions
        elif data is None:
            self.permissions = set()
        else:
            raise TypeError(f'Cannot parse in a {type(data)}!')

    def set(self, flag):
        '''Set a specific flag.'''

    def clear(self, flag):
        '''Clear a specific flag.'''

    def parse_str(self, data):
        '''Parse a string as a permission.'''
        data = int(data)
        self.permissions = set()
        for permission in self.PermissionFlags:
            if permission & data:
                self.permissions.add(permission)

    def __and__(self, other):
        '''Logically AND two Permissions together.'''
        if type(other) is Permissions:
            new_permissions = Permissions()

            for permission in other.permissions:
                if permission in self.permissions:
                    new_permissions.set(permission)
            for permission in self.permissions:
                if permission in other.permissions:
                    new_permissions.set(permission)
            return new_permissions

        elif type(other) is self.PermissionFlags:
            return other in self.permissions

        elif type(other) in [int, str]:
            if type(other) is str:
                other = int(other)

        raise TypeError(f'Cannot AND a Permission against a {type(other)}!')

    def __rand__(self, other):
        '''See __and__().'''
        return self.__and__(other)

    def __eq__(self, other):
        '''Determine if two permission objects are equivalent.'''
        assert type(other) is Permissions
        return self.permissions == other.permissions



import emoji

from.base_object import BaseDiscordObject


class Emoji(BaseDiscordObject):
    '''Emojis.'''

    def __init__(self, name: str = None, unicode: str = None):
        '''Initalize Emoji.

        TODO: Document parameters.
        '''
        self.name = name
        self.unicode = unicode

        if (unicode is None) and (name is not None):
            assert type(self.name) is str
            self.unicode = emoji.unicode_codes.EMOJI_UNICODE_ENGLISH[self.name]

    def __str__(self):
        '''Return string representation.'''
        return f'Emoji(name={self.name}, unicode={self.unicode})'



import abc
import datetime
from typing import Optional, List, Dict

import validators  # type: ignore

from .base_object import BaseDiscordObject
from . import enumerations


class Embed(BaseDiscordObject):
    '''Rich context holder for a message.'''

    EMBED_TYPES = enumerations.EMBED_TYPES

    title: Optional[str] = None
    type: Optional['enumerations.EMBED_TYPES'] = None  # type: ignore # DEPRECATED
    description: Optional[str] = None  # type: ignore # description of embed
    url: Optional[str] = None  # type: ignore # url of embed
    timestamp: Optional['datetime.datetime'] = None  # type: ignore # timestamp	timestamp of embed content
    color: Optional[int] = None  # type: ignore # color code of the embed
    footer: Optional['EmbedFooter'] = None  # type: ignore # footer information
    image: Optional['EmbedImage'] = None  # type: ignore # image information
    thumbnail: Optional['EmbedThumbnail'] = None  # type: ignore # thumbnail information
    video: Optional['EmbedVideo'] = None  # type: ignore # video information
    provider: Optional['EmbedProvider'] = None  # type: ignore # provider information
    author: Optional['EmbedAuthor'] = None  # type: ignore # author information
    fields: Optional[List['EmbedField']] = None  # type: ignore # fields information

    def __str__(self):
        '''Return string representation.'''
        return 'Embed()'

    def to_dict(self) -> 'dict':  # noqa: C901
        '''Convert object to dictionary suitable for API or other generic useage.'''
        new_dict: Dict[str, object] = dict()

        if (self.title is not None):
            new_dict['title'] = self.title
        if (self.type is not None) and (type(self.type) is enumerations.EMBED_TYPES):
            new_dict['type'] = self.type.name
        if (self.description is not None):
            new_dict['description'] = self.description
        if (self.url is not None):
            new_dict['url'] = self.url
        if (self.timestamp is not None) and (type(self.timestamp) is datetime.datetime):
            new_dict['timestamp'] = self.timestamp.isoformat()
        if (self.color is not None):
            new_dict['color'] = self.color
        if (self.footer is not None) and (type(self.footer) is EmbedFooter):
            new_dict['footer'] = self.footer.to_dict()
        if (self.image is not None) and (type(self.image) is EmbedImage):
            new_dict['image'] = self.image.to_dict()
        if (self.thumbnail is not None) and (type(self.thumbnail) is EmbedThumbnail):
            new_dict['thumbnail'] = self.thumbnail.to_dict()
        if (self.video is not None) and (type(self.video) is EmbedVideo):
            new_dict['video'] = self.video.to_dict()
        if (self.provider is not None) and (type(self.provider) is EmbedProvider):
            new_dict['provider'] = self.provider.to_dict()
        if (self.author is not None) and (type(self.author) is EmbedAuthor):
            new_dict['author'] = self.author.to_dict()
        if (self.fields is not None) and (type(self.fields) is list):
            new_dict['fields'] = list()
            assert type(new_dict['fields']) is list
            for field in self.fields:
                if type(field) is EmbedField:
                    new_dict['fields'].append(field.to_dict())

        return new_dict

    def generate(self,
                 title: Optional[str] = None,
                 type: Optional['enumerations.EMBED_TYPES'] = None,
                 description: Optional[str] = None,
                 url: Optional[str] = None,
                 timestamp: Optional['datetime.datetime'] = None,
                 color: Optional[int] = None,  # TODO: Support a rich color class here.
                 ):
        '''Generate various elements of an Embed.'''
        if title is not None:
            self.title = title

        if type is not None:
            self.type = type

        if description is not None:
            self.description = description

        if url is not None:
            self.url = url

        if timestamp is not None:
            self.timestamp = timestamp

        if color is not None:
            self.color = color

    def validate(self):  # noqa: C901
        '''Validate object is prepared for dispatch to discord.'''
        total_length = 0

        if self.title is not None:
            assert type(self.title) is str,\
                f'Got invalid type {type(self.title)} for title.'
            assert len(self.title) <= 256,\
                f'God invalid length of title, {len(self.title)} characters long. Max is 256.'
            total_length += len(self.title)

        if self.type is not None:
            assert type(self.type) is enumerations.EMBED_TYPES,\
                f'Got invalid type {type(self.type)} for type.'

        if self.description is not None:
            assert type(self.description) is str,\
                f'Got invalid type {type(self.description)} for description.'
            assert len(self.description) <= 4096,\
                f'Got invalid length of description, {len(self.description)} characters long. Max is 4096.'
            total_length += len(self.description)

        if self.url is not None:
            assert type(self.url) is str,\
                f'Got invalid type {type(self.url)} for url.'
            if not validators.url(self.url, public=False):
                raise AssertionError(f'URL fails validation: [{self.url}].')

        if self.timestamp is not None:
            assert type(self.timestamp) is datetime.datetime,\
                f'Got invalid type {type(self.timestamp)} for timestamp.'

        if self.color is not None:
            assert type(self.color) is int,\
                f'Got invalid type {type(self.color)} for timestamp.'

        if self.footer is not None:
            assert type(self.footer) is EmbedFooter,\
                f'Got invalid type {type(self.footer)} for footer. Must be EmbedFooter.'
            self.footer.validate()
            total_length += len(self.footer.text)

        if self.fields is not None:
            assert type(self.fields) is list,\
                f'Got invalid type {type(self.fields)} for fields.'
            assert len(self.fields) <= 25,\
                f'Got invalid length of fields, {len(self.fields)} elements long. Max is 25.'
            for field in self.fields:
                field.validate()
                total_length += len(field.name)
                total_length += len(field.value)

        if self.author is not None:
            assert type(self.author) is EmbedAuthor,\
                f'Got invalid type {type(self.author)} for author.'
            self.author.validate()

        if self.image is not None:
            assert type(self.image) is EmbedImage,\
                f'Got invalid type {type(self.image)} for image.'
            self.image.validate()

        if self.thumbnail is not None:
            assert type(self.thumbnail) is EmbedThumbnail,\
                f'Got invalid type {type(self.thumbnail)} for thumbnail.'
            self.thumbnail.validate()

        if self.video is not None:
            assert type(self.video) is EmbedVideo,\
                f'Got invalid type {type(self.video)} for video.'
            self.video.validate()

        if self.provider is not None:
            assert type(self.provider) is EmbedProvider,\
                f'Got invalid type {type(self.provider)} for provider.'
            self.provider.validate()

        assert total_length <= 6000,\
            f'Total characters in embed is {total_length}. Max API limit is 6000.'

    def add_footer(self,
                   text: str,
                   icon_url: str = None,
                   proxy_icon_url: str = None,
                   ) -> 'EmbedFooter':
        '''Add footer to the Embed.'''
        self.footer = EmbedFooter()
        self.footer.text = text
        if icon_url is not None:
            self.footer.icon_url = icon_url
        if proxy_icon_url is not None:
            self.footer.proxy_icon_url = proxy_icon_url
        return self.footer

    def add_image(self,
                  url: str,
                  proxy_url: str = None,
                  height: int = None,
                  width: int = None,
                  ) -> 'EmbedImage':
        '''Add image to the Embed.'''
        self.image = EmbedImage()
        self.image.url = url
        if proxy_url is not None:
            self.image.proxy_url = proxy_url
        if height is not None:
            self.image.height = height
        if width is not None:
            self.image.width = width
        return self.image

    def add_thumbnail(self,
                      url: str,
                      proxy_url: str = None,
                      height: int = None,
                      width: int = None,
                      ) -> 'EmbedThumbnail':
        '''Add thumbnail to the Embed.'''
        self.thumbnail = EmbedThumbnail()
        self.thumbnail.url = url
        if proxy_url is not None:
            self.thumbnail.proxy_url = proxy_url
        if height is not None:
            self.thumbnail.height = height
        if width is not None:
            self.thumbnail.width = width
        return self.thumbnail

    def add_video(self,
                  url: str,
                  proxy_url: str = None,
                  height: int = None,
                  width: int = None,
                  ) -> 'EmbedVideo':
        '''Add video to the Embed.'''
        self.video = EmbedVideo()
        self.video.url = url
        if proxy_url is not None:
            self.video.proxy_url = proxy_url
        if height is not None:
            self.video.height = height
        if width is not None:
            self.video.width = width
        return self.video

    def add_provider(self,
                     name: Optional[str] = None,
                     url: Optional[str] = None,
                     ) -> 'EmbedProvider':
        '''Add provider to the Embed.'''
        self.provider = EmbedProvider()
        if name is not None:
            self.provider.name = name
        if url is not None:
            self.provider.url = url
        return self.provider

    def add_author(self,
                   name: str,
                   url: Optional[str] = None,
                   icon_url: Optional[str] = None,
                   proxy_icon_url: Optional[str] = None,
                   ) -> 'EmbedAuthor':
        '''Add author to the Embed.'''
        self.author = EmbedAuthor()
        self.author.name = name
        if url is not None:
            self.author.url = url
        if icon_url is not None:
            self.author.icon_url = icon_url
        if proxy_icon_url is not None:
            self.author.proxy_icon_url = proxy_icon_url
        return self.author

    def add_field(self,
                  name: str,
                  value: str,
                  inline: bool = False,
                  ) -> 'EmbedField':
        '''Add a generic field to the Embed.'''
        if not isinstance(self.fields, list):
            self.fields = list()
        new_field = EmbedField()
        new_field.name = name
        new_field.value = value
        new_field.inline = inline
        self.fields.append(new_field)
        return new_field


class EmbedFooter(BaseDiscordObject):
    '''Footer details for the Embed.'''

    text: str = None  # type: ignore  # footer text
    icon_url: Optional[str] = None  # url of footer icon (only supports http(s) and attachments)
    proxy_icon_url: Optional[str] = None  # a proxied url of footer icon

    def to_dict(self) -> 'dict':
        '''Convert object to dictionary suitable for API or other generic useage.'''
        new_dict: Dict[str, object] = dict()
        if self.text is not None:
            new_dict['text'] = self.text
        if self.icon_url is not None:
            new_dict['icon_url'] = self.icon_url
        if self.proxy_icon_url is not None:
            new_dict['proxy_icon_url'] = self.proxy_icon_url
        return new_dict

    def validate(self):
        '''Validate object is prepared for dispatch to discord.'''
        assert len(self.text) <= 2048,\
            f'Got invalid length of text, {len(self.text)} elements long. Max is 2048.'

        if self.icon_url is not None:
            assert type(self.icon_url) is str,\
                f'Got invalid type {type(self.icon_url)} for icon_url. Must be str.'
            assert self.icon_url.startswith(('http://', 'https://')),\
                f'Got invalid icon_url {self.icon_url}, must start with http:// or https://.'

        if self.proxy_icon_url is not None:
            assert type(self.proxy_icon_url) is str,\
                f'Got invalid type {type(self.proxy_icon_url)} for proxy_icon_url. Must be str.'
            assert self.proxy_icon_url.startswith(('http://', 'https://')),\
                f'Got invalid proxy_icon_url {self.proxy_icon_url}, must start with http:// or https://.'


class EmbedImage(BaseDiscordObject):
    '''Image to embed.'''

    url: str = None  # type: ignore  # source url of thumbnail (only supports http(s) and attachments)
    proxy_url: Optional[str] = None  # a proxied url of the thumbnail
    height: Optional[int] = None  # height of thumbnail
    width: Optional[int] = None  # width of thumbnail

    def to_dict(self) -> 'dict':
        '''Convert object to dictionary suitable for API or other generic useage.'''
        new_dict: Dict[str, object] = dict()
        new_dict['url'] = self.url

        if self.proxy_url is not None:
            new_dict['proxy_url'] = self.proxy_url

        if self.height is not None:
            new_dict['height'] = self.height

        if self.width is not None:
            new_dict['width'] = self.width

        return new_dict

    def validate(self):
        '''Validate object is prepared for dispatch to discord.'''
        assert type(self.url) is str,\
            f'Got invalid type {type(self.url)} for url. Must be str.'
        assert self.url.startswith(('http://', 'https://')),\
            f'Got invalid url {self.url}, must start with http:// or https://.'

        if self.proxy_url is not None:
            assert type(self.proxy_url) is str,\
                f'Got invalid type {type(self.proxy_url)} for proxy_url. Must be str.'
            assert self.proxy_url.startswith(('http://', 'https://')),\
                f'Got invalid proxy {self.proxy_url}, must start with http:// or https://.'

        if self.height is not None:
            assert type(self.height) is int,\
                f'Got invalid type {type(self.height)} for height. Must be int.'

        if self.width is not None:
            assert type(self.width) is int,\
                f'Got invalid type {type(self.width)} for url. Must be int.'


class EmbedThumbnail(BaseDiscordObject):
    '''Thumbnail to embed.'''

    url: str = None  # type: ignore  # source url of thumbnail (only supports http(s) and attachments)
    proxy_url: Optional[str] = None  # a proxied url of the thumbnail
    height: Optional[int] = None  # height of thumbnail
    width: Optional[int] = None  # width of thumbnail

    def to_dict(self) -> 'dict':
        '''Convert object to dictionary suitable for API or other generic useage.'''
        new_dict: Dict[str, object] = dict()
        new_dict['url'] = self.url

        if self.proxy_url is not None:
            new_dict['proxy_url'] = self.proxy_url

        if self.height is not None:
            new_dict['height'] = self.height

        if self.width is not None:
            new_dict['width'] = self.width

        return new_dict

    def validate(self):
        '''Validate object is prepared for dispatch to discord.'''
        assert type(self.url) is str,\
            f'Got invalid type {type(self.url)} for url. Must be str.'
        assert self.url.startswith(('http://', 'https://')),\
            f'Got invalid url {self.url}, must start with http:// or https://.'

        if self.proxy_url is not None:
            assert type(self.proxy_url) is str,\
                f'Got invalid type {type(self.proxy_url)} for proxy_url. Must be str.'
            assert self.proxy_url.startswith(('http://', 'https://')),\
                f'Got invalid proxy_url {self.proxy_url}, must start with http:// or https://.'

        if self.height is not None:
            assert type(self.height) is int,\
                f'Got invalid type {type(self.height)} for height. Must be int.'

        if self.width is not None:
            assert type(self.width) is int,\
                f'Got invalid type {type(self.width)} for url. Must be int.'


class EmbedVideo(BaseDiscordObject):
    '''Video to embed.'''

    url: Optional[str] = None  # source url of thumbnail (only supports http(s) and attachments)
    proxy_url: Optional[str] = None  # a proxied url of the thumbnail
    height: Optional[int] = None  # height of thumbnail
    width: Optional[int] = None  # width of thumbnail

    def to_dict(self) -> 'dict':
        '''Convert object to dictionary suitable for API or other generic useage.'''
        new_dict: Dict[str, object] = dict()
        if self.url is not None:
            new_dict['url'] = self.url

        if self.proxy_url is not None:
            new_dict['proxy_url'] = self.proxy_url

        if self.height is not None:
            new_dict['height'] = self.height

        if self.width is not None:
            new_dict['width'] = self.width

        return new_dict

    def validate(self):
        '''Validate object is prepared for dispatch to discord.'''
        if self.url is not None:
            assert type(self.url) is str,\
                f'Got invalid type {type(self.url)} for url. Must be str.'
            assert self.url.startswith(('http://', 'https://')),\
                f'Got invalid url {self.url}, must start with http:// or https://.'

        if self.proxy_url is not None:
            assert type(self.proxy_url) is str,\
                f'Got invalid type {type(self.proxy_url)} for proxy_url. Must be str.'
            assert self.proxy_url.startswith(('http://', 'https://')),\
                f'Got invalid proxy_url {self.proxy_url}, must start with http:// or https://.'

        if self.height is not None:
            assert type(self.height) is int,\
                f'Got invalid type {type(self.height)} for height. Must be int.'

        if self.width is not None:
            assert type(self.width) is int,\
                f'Got invalid type {type(self.width)} for url. Must be int.'


class EmbedProvider(BaseDiscordObject):
    '''Provider of the content from the Embed.'''

    name: str = None  # type: ignore  # name of author
    url: str = None  # type: ignore  # url of author

    def to_dict(self) -> 'dict':
        '''Convert object to dictionary suitable for API or other generic useage.'''
        new_dict: Dict[str, object] = dict()

        if self.name is not None:
            new_dict['name'] = self.name

        if self.url is not None:
            new_dict['url'] = self.url
        return new_dict

    def validate(self):
        '''Validate object is prepared for dispatch to discord.'''
        if self.name is not None:
            assert type(self.name) is str,\
                f'Got invalid type {type(self.name)} for name. Must be str.'
            assert len(self.name) <= 256,\
                f'Got invalid length of name, {len(self.name)} characters long. Max is 256.'

        if self.url is not None:
            assert type(self.url) is str,\
                f'Got invalid type {type(self.url)} for url. Must be str.'
            assert self.url.startswith(('http://', 'https://')),\
                f'Got invalid url {self.url}, must start with http:// or https://.'


class EmbedAuthor(BaseDiscordObject):
    '''Author of the content from the Embed.'''

    name: str = None  # type: ignore  # name of author
    url: str = None  # type: ignore  # url of author
    icon_url: str = None  # type: ignore  # url of author icon (only supports http(s) and attachments)
    proxy_icon_url: str = None  # type: ignore  # a proxied url of author icon

    def to_dict(self) -> 'dict':
        '''Convert object to dictionary suitable for API or other generic useage.'''
        new_dict: Dict[str, object] = dict()

        if self.name is not None:
            new_dict['name'] = self.name

        if self.url is not None:
            new_dict['url'] = self.url

        if self.icon_url is not None:
            new_dict['icon_url'] = self.icon_url

        if self.proxy_icon_url is not None:
            new_dict['proxy_icon_url'] = self.proxy_icon_url

        return new_dict

    def validate(self):
        '''Validate object is prepared for dispatch to discord.'''
        if self.name is not None:
            assert type(self.name) is str,\
                f'Got invalid type {type(self.name)} for name. Must be str.'
            assert len(self.name) <= 256,\
                f'Got invalid length of name, {len(self.name)} characters long. Max is 256.'

        if self.url is not None:
            assert type(self.url) is str,\
                f'Got invalid type {type(self.url)} for url. Must be str.'
            assert self.url.startswith(('http://', 'https://')),\
                f'Got invalid url {self.url}, must start with http:// or https://.'

        if self.icon_url is not None:
            assert type(self.icon_url) is str,\
                f'Got invalid type {type(self.icon_url)} for icon_url. Must be str.'
            assert self.icon_url.startswith(('http://', 'https://')),\
                f'Got invalid icon_url {self.icon_url}, must start with http:// or https://.'

        if self.proxy_icon_url is not None:
            assert type(self.proxy_icon_url) is str,\
                f'Got invalid type {type(self.proxy_icon_url)} for proxy_icon_url. Must be str.'
            assert self.proxy_icon_url.startswith(('http://', 'https://')),\
                f'Got invalid proxy_icon_url {self.proxy_icon_url}, must start with http:// or https://.'


class EmbedField(BaseDiscordObject):
    '''Single field of an Embed.'''

    name: str = None  # type: ignore # name of the field
    value: str = None  # type: ignore # value of the field
    inline: bool = None  # type: ignore # whether or not this field should display inline

    def to_dict(self) -> 'dict':
        '''Convert object to dictionary suitable for API or other generic useage.'''
        new_dict: Dict[str, object] = dict()

        if self.name is not None:
            new_dict['name'] = self.name

        if self.value is not None:
            new_dict['value'] = self.value

        if self.inline is not None:
            new_dict['inline'] = self.inline

        return new_dict

    def validate(self):
        '''Validate object is prepared for dispatch to discord.'''
        if self.name is not None:
            assert type(self.name) is str,\
                f'Got invalid type {type(self.name)} for name. Must be str.'
            assert len(self.name) <= 256,\
                f'Got invalid length of name, {len(self.name)} characters long. Max is 256.'

        if self.value is not None:
            assert type(self.value) is str,\
                f'Got invalid type {type(self.value)} for value. Must be str.'
            assert len(self.value) <= 1024,\
                f'Got invalid length of value, {len(self.value)} characters long. Max is 1024.'

        if self.inline is not None:
            assert type(self.inline) is bool,\
                f'Got invalid type {type(self.inline)} for inline. Must be bool.'


class EmbedAdder(abc.ABC):
    '''Allow other objects to start adding components to themselves with a common set of helper functions.

    Caution: This is an abstract class, and is not intended for direct instantiation.
    '''

    embeds: Optional[List['Embed']] = None

    def add_embeds(self) -> 'Embed':
        '''Add embeds to the object.'''
        if not isinstance(self.embeds, list):
            self.embeds: Optional[List['Embed']] = list()
        assert type(self.embeds) is list
        new_embed = Embed()
        self.embeds.append(new_embed)

        return new_embed



import asyncio
from collections import defaultdict
from typing import Any, Optional, Union, List, Dict

from ...client import api

from ..base_object import BaseDiscordObject

from .. import channel
from .. import snowflake

from .. import user as ext_user, message as ext_message, embed as ext_embed, role

from . import enumerations, components as ext_components, command


class Interaction(BaseDiscordObject):
    '''Response given from the server after an user activated an interaction of some type.

    Attributes:
        id (Snowflake): ID of the specific interaction.
        application_id (Snowflake): ID of the application that registered the Interaction.
        type (INTERACTION_TYPES): The type of interaction.
        data (InteractionData): Data generated by the interaction.
        guild_id (Snowflake|None): ID of the guild the command was generated from. This still exists for commands registered to the Global space!
        channel_id (Snowflake|None): ID of the channel the command was generated from.
        member (Member|None): Member that activated the command.
        user (User|None): User that activated the command.
        token (str): Unique one time use token used in an InteractionResponse.
        version (int): Unique version number generated by discord when the command is registered.
        message (Message|None): For components, the message they were attached to
    '''

    INTERACTION_RESPONSE_TYPES = enumerations.INTERACTION_RESPONSE_TYPES

    id: snowflake.Snowflake
    application_id: snowflake.Snowflake
    type: enumerations.INTERACTION_TYPES
    data: Optional['InteractionData']
    guild_id: Optional[snowflake.Snowflake]
    # TODO: add a Guild object from the ID.
    channel_id: Optional[snowflake.Snowflake]
    # TODO: add a Channel object from the ID.
    member: Optional['ext_user.Member']
    user: Optional['ext_user.User']
    token: str
    version: int
    message: Optional['ext_message.Message']

    def __init__(self):
        '''Simple initialization.'''
        self._response_generated = False
        self.id = None

    @property
    def can_respond(self):
        '''Determine if you can respond to this interaction.'''
        # TODO: We could check timestamps here as well!
        return self._response_generated is False

    @property
    def can_followup(self):
        '''Determine if you can followup to this interaction.'''
        # TODO: We could check timestamps here as well!
        return self._response_generated is True

    def from_dict(self, data: dict) -> 'Interaction':
        '''Import data from dict and populate object with it.'''
        self._log.info('Parsing  a Interaction dict')
        self.application_id = snowflake.Snowflake(data['application_id'])
        self.id = snowflake.Snowflake(data['id'])
        self.token = str(data['token'])
        self.version = int(data['version'])
        self.type = enumerations.INTERACTION_TYPES(data['type'])
        if 'channel_id' in data:
            self.channel_id = snowflake.Snowflake(data['channel_id'])
        if 'guild_id' in data:
            self.guild_id = snowflake.Snowflake(data['guild_id'])
        if 'data' in data:
            guild_id = snowflake.Snowflake(data['guild_id']) if 'guild_id' in data else None
            self.data = InteractionData().from_dict(data['data'], guild_id)
        if 'channel_id' in data:
            self.channel_id = snowflake.Snowflake(data['channel_id'])
        if 'user' in data:
            self.user = ext_user.User().from_dict(data['user'])
        if 'member' in data:
            self.member = ext_user.Member().from_dict(data['member'])
        if 'message' in data:
            self.message = ext_message.Message().from_dict(data['message'])

        return self

    def generate_response(self,
                          type: enumerations.INTERACTION_RESPONSE_TYPES = enumerations.INTERACTION_RESPONSE_TYPES.CHANNEL_MESSAGE_WITH_SOURCE,
                          ephemeral: bool = False,
                          ) -> 'InteractionResponse':
        '''Produce a InteractionResponse object to follow up against the Interaction.

        Arguments:
            type (INTERACTION_RESPONSE_TYPES): Type of response to give.
            ephemeral (bool): Only user should see the response.
        '''
        if self._response_generated:
            raise RuntimeError('You cannot reuse a token from a response. Call generate_followup()')
        else:
            self._response_generated = True
        new_response = InteractionResponse()
        new_response.interaction_id = self.id
        new_response.interaction_token = self.token
        new_response.type = type
        # new_response.generate(content, tts)  # TODO: Support this here too!
        if ephemeral:
            new_response.data.flags |= enumerations.INTERACTION_CALLBACK_FLAGS.EPHEMERAL
        return new_response

    def generate_followup(self,
                          content: Optional[str] = None,
                          tts: Optional[bool] = None,
                          ) -> 'InteractionFollowup':
        '''Produce a InteractionFollowup object to follow up against the Interaction.'''
        if self._response_generated is False:
            raise RuntimeError('You cannot followup on an interaction until you respond to it. Call generate_response()')
        new_followup = InteractionFollowup()
        new_followup.interaction_id = self.id
        new_followup.interaction_token = self.token
        new_followup.generate(content, tts)
        return new_followup


class InteractionData(BaseDiscordObject):
    '''Handle data from an Interaction.'''

    id: snowflake.Snowflake  # snowflake the ID of the invoked command Application Command
    name: str  # string the name of the invoked command Application Command
    type: enumerations.COMMAND_TYPE  # integer the type of the invoked command Application Command
    resolved: Dict[str, dict]  # ? resolved data converted users + roles + channels Application Command
    options: Dict[str, Any]  # ? array of application command interaction data option the params + values from the user Application Command
    custom_id: str  # ? string the custom_id of the component Component
    component_type: Optional[enumerations.COMPONENT_TYPES]  # ? integer the type of the component Component
    values: Optional[List['command.CommandOptions']]  # ? array of select option values the values the user selected Component (Select)
    target_id: Optional[snowflake.Snowflake]  # ? snowflake id the of user or message targetted by a user or message command

    def from_dict(self, data: dict, guild_id: Optional['snowflake.Snowflake'] = None) -> 'InteractionData':  # noqa: C901
        '''Import data from dict and populate object with it.'''
        self._log.info('Parse a InteractionData dict.')
        if 'id' in data:
            self.id = snowflake.Snowflake(data['id'])
        if 'name' in data:
            self.name = data['name']
        if 'type' in data:
            self.type = enumerations.COMMAND_TYPE(data['type'])
        if 'resolved' in data:
            self._log.info('FOUND RESOLVED!')
            self.resolved = defaultdict(lambda: dict())
            for resolution_type in data['resolved']:
                self._log.info(f'Resolved [{resolution_type}].')
                for entry_id in data['resolved'][resolution_type]:
                    if resolution_type == 'members':
                        self.resolved[resolution_type][entry_id] = ext_user.Member()
                    elif resolution_type == 'users':
                        self.resolved[resolution_type][entry_id] = ext_user.User()
                    elif resolution_type == 'messages':
                        self.resolved[resolution_type][entry_id] = ext_message.Message()
                    elif resolution_type == 'channels':
                        self.resolved[resolution_type][entry_id] = channel.ChannelImporter()
                    elif resolution_type == 'roles':
                        self.resolved[resolution_type][entry_id] = role.Role()
                    else:
                        self._log.critical(f'Cannot resolve type [{resolution_type}]!')
                        raise TypeError

                    self.resolved[resolution_type][entry_id].from_dict(data['resolved'][resolution_type][entry_id])
        if 'component_type' in data:
            self.component_type = enumerations.COMPONENT_TYPES(data['component_type'])
        if 'custom_id' in data:
            self.custom_id = str(data['custom_id'])
        if 'target_id' in data:
            self.target_id = snowflake.Snowflake(data['target_id'])

        self.options = dict()
        if 'options' in data:
            for option_dict in data['options']:
                self.options[option_dict['name']] = InteractionDataOptionStructure().from_dict(option_dict, guild_id).parse(guild_id)
        return self


class InteractionDataOptionStructure(BaseDiscordObject):
    '''InteractionDataOptionStructure.'''

    name: str                                                  # string the name of the invoked command Application Command
    type: enumerations.COMMAND_OPTION                          # integer the type of the invoked command Application Command
    value: Optional[Union[str, int, bool, 'snowflake.Snowflake', float]]               # the value of the pair
    options: Optional[Dict[str, Any]]  # Present when command is a group or subcommand

    def __getitem__(self, key: str):
        '''Return item from the options dict.'''
        if hasattr(self, 'options') and type(self.options) is dict:
            return self.options[key]

    def __contains__(self, item: str):
        '''Determine if item is in the options dict.'''
        if hasattr(self, 'options') and type(self.options) is dict:
            return item in self.options

    def from_dict(self, data: dict, guild_id: Optional['snowflake.Snowflake'] = None) -> 'InteractionDataOptionStructure':  # noqa: C901
        '''Import data from dict and populate object with it.'''
        self.name = data['name']
        self.type = enumerations.COMMAND_OPTION(data['type'])
        if 'value' in data:
            if self.type == enumerations.COMMAND_OPTION.SUB_COMMAND:
                raise ValueError('Sub Commands should not have values!')
            elif self.type == enumerations.COMMAND_OPTION.SUB_COMMAND_GROUP:
                raise ValueError('Sub Command Groups should not have values!')
            elif self.type in [enumerations.COMMAND_OPTION.STRING, enumerations.COMMAND_OPTION.INTEGER, enumerations.COMMAND_OPTION.BOOLEAN, enumerations.COMMAND_OPTION.NUMBER]:
                self.value = data['value']
            elif self.type == enumerations.COMMAND_OPTION.USER:
                # TODO: Actually lookup the user here and replace the snowflake with a full object.
                self.value = snowflake.Snowflake(data['value'])
            elif self.type == enumerations.COMMAND_OPTION.CHANNEL:
                # TODO: Actually lookup the channel here and replace the snowflake with a full object.
                self.value = snowflake.Snowflake(data['value'])
            elif self.type == enumerations.COMMAND_OPTION.ROLE:
                # TODO: Actually lookup the role here and replace the snowflake with a full object.
                self.value = snowflake.Snowflake(data['value'])
            elif self.type == enumerations.COMMAND_OPTION.MENTIONABLE:
                # TODO: Actually lookup the mentionable here and replace the snowflake with a full object.
                self.value = snowflake.Snowflake(data['value'])
        self.options = dict()
        if 'options' in data:
            for option_dict in data['options']:
                self.options[option_dict['name']] = InteractionDataOptionStructure().from_dict(option_dict, guild_id).parse(guild_id)
        return self

    def parse(self, guild_id: Optional['snowflake.Snowflake'] = None):  # noqa: C901
        '''Look at the type field of self and attempt to return a sane result.

        For SUB_COMMAND and SUB_COMMAND_GROUP types, return another InteractionDataOptionStructure object.
        For all others, attempt to reference the type and handle it.
        '''
        CO = enumerations.COMMAND_OPTION

        if self.type in [CO.SUB_COMMAND, CO.SUB_COMMAND_GROUP]:
            return self

        if self.type in [CO.STRING, CO.INTEGER, CO.BOOLEAN, CO.NUMBER]:
            return self.value

        assert type(self.value) is snowflake.Snowflake

        if self.type == CO.USER:
            return ext_user.User().from_dict(asyncio.run(api.API.get_user(self.value)))

        if self.type == CO.CHANNEL:
            return channel.Channel().from_dict(asyncio.run(api.API.get_channel(self.value)))

        if self.type == CO.ROLE:
            assert type(guild_id) is snowflake.Snowflake
            roles_list = asyncio.run(api.API.get_guild_roles(guild_id))
            for role_dict in roles_list:
                if role_dict['id'] == self.value:
                    return role.Role().from_dict(role_dict)

        if self.type == CO.MENTIONABLE:
            try:
                return ext_user.User().from_dict(asyncio.run(api.API.get_user(self.value)))
            except Exception:
                assert type(guild_id) is snowflake.Snowflake
                roles_list = asyncio.run(api.API.get_guild_roles(guild_id))
                for role_dict in roles_list:
                    if role_dict['id'] == self.value:
                        return role.Role().from_dict(role_dict)


class InteractionResponse(BaseDiscordObject):
    '''Manage a response message to an interaction.'''

    INTERACTION_RESPONSE_TYPES = enumerations.INTERACTION_RESPONSE_TYPES

    type: enumerations.INTERACTION_RESPONSE_TYPES
    data: 'InteractionCallback'

    # The following are part of our local book keeping, not the discord structure.
    interaction_id: 'snowflake.Snowflake'
    interaction_token: str

    def __init__(self):
        '''Initialize InteractionResponse.'''
        self.data = InteractionCallback()
        self.last_followup_message: Optional['ext_message.Message'] = None

    def to_dict(self) -> dict:
        '''Convert object to dictionary suitable for API or other generic useage.'''
        new_dict: Dict[str, object] = dict()
        new_dict['type'] = self.type.value
        new_dict['data'] = self.data.to_dict()
        return new_dict

    async def send(self) -> None:
        '''Send response.'''
        await api.API.create_interaction_response(self.interaction_id, self.interaction_token, self.to_dict())

    def generate(self,
                 content: Optional[str] = None,
                 tts: Optional[bool] = None,
                 ephemeral: Optional[bool] = None,
                 ):
        '''Passthrough to data generate function.'''
        return self.data.generate(tts, content, ephemeral)

    def add_components(self) -> 'ext_components.ActionRow':
        '''Add component objects to the data attribute.'''
        return self.data.add_components()

    def add_embeds(self) -> 'ext_embed.Embed':
        '''Add embed objects to the data attribute.'''
        return self.data.add_embeds()


class InteractionFollowup(BaseDiscordObject):
    '''Manage a followup message to an interaction.'''

    data: 'InteractionCallback'

    # The following are part of our local book keeping, not the discord structure.
    interaction_id: 'snowflake.Snowflake'
    interaction_token: str

    def __init__(self):
        '''Initialize InteractionFollowup.'''
        self.data = InteractionCallback()
        self.last_followup_message: Optional['ext_message.Message'] = None

    def to_dict(self) -> dict:
        '''Convert object to dictionary suitable for API or other generic useage.'''
        new_dict: Dict[str, object] = dict()
        new_dict['data'] = self.data.to_dict()
        return new_dict

    async def send(self) -> 'ext_message.Message':
        '''Send followup.'''
        # TODO: We can do some cool stuff with overrides here, look into that. https://discord.com/developers/docs/resources/webhook#execute-webhook
        data = await self._generate_webhook_data()
        msg_dict = await api.API.create_followup_message(self.interaction_token, data)
        new_message = ext_message.Message().from_dict(msg_dict)
        self.last_followup_message = new_message
        return new_message

    async def edit_original_response(self):
        '''Edit the initial response to the original interaction.'''
        data = await self._generate_webhook_data()
        await api.API.edit_original_interaction_response(self.interaction_token, data)

    async def delete_initial_response(self):
        '''Delete the initial response to the original interaction.'''
        await api.API.delete_original_interaction_response(self.interaction_token)

    async def edit_followup_message(self, message: Optional['ext_message.Message'] = None) -> 'ext_message.Message':
        '''Apply modifications to the given message object.'''
        # TODO: We can do some cool stuff with overrides here, look into that. https://discord.com/developers/docs/resources/webhook#execute-webhook
        data = await self._generate_webhook_data()
        if message is None:
            if self.last_followup_message is not None:
                message = self.last_followup_message
            else:
                raise ValueError('Must give specific followup message, or have sent one already.')
        msg_dict = await api.API.edit_followup_message(self.interaction_token, message.id, data)
        new_message = ext_message.Message().from_dict(msg_dict)
        self.last_followup_message = new_message
        return new_message

    async def delete_followup_message(self, message: Optional['ext_message.Message'] = None) -> None:
        '''Delete the given followup message.

        If a message was just send using the send() method of this object, and no message argument is given, then attempt to delete the last followup message.
        '''
        if message is None:
            if self.last_followup_message is not None:
                message = self.last_followup_message
            else:
                raise ValueError('Must give specific followup message, or have sent one already.')
        await api.API.delete_followup_message(self.interaction_token, message.id)

    async def _generate_webhook_data(self) -> dict:
        data_structure: dict = dict()
        if hasattr(self.data, 'content'):
            data_structure['content'] = self.data.content
        if hasattr(self.data, 'embeds'):
            data_structure['embeds'] = list()
            assert type(self.data.embeds) is list
            for embed in self.data.embeds:
                data_structure['embeds'].append(embed.to_dict())
        if hasattr(self.data, 'components'):
            data_structure['components'] = list()
            assert type(self.data.components) is list
            for component in self.data.components:
                data_structure['components'].append(component.to_dict())
        return data_structure

    def generate(self,
                 content: Optional[str] = None,
                 tts: Optional[bool] = None,
                 ):
        '''Customize common attributes of a callback.

        Arguments:
            tts (bool): Send message as text to speech.
            content (str): Text contents of the message to send.

        Note:
            You may be tempted to grab the resulting `InteractionCallback` and adjust the `ephemeral` flag. Followup messages do not support this!
        '''
        return self.data.generate(tts, content, None)

    def add_components(self) -> 'ext_components.ActionRow':
        '''Add components objects to the data attribute.'''
        return self.data.add_components()

    def add_embeds(self) -> 'ext_embed.Embed':
        '''Add embed objects to the data attribute.'''
        return self.data.add_embeds()


class InteractionCallback(BaseDiscordObject, ext_components.ComponentAdder, ext_embed.EmbedAdder):
    '''InteractionCallback.'''

    INTERACTION_CALLBACK_FLAGS = enumerations.INTERACTION_CALLBACK_FLAGS

    tts: Optional[bool]
    content: Optional[str]
    embeds: Optional[List['ext_embed.Embed']]  # TODO: Support embeds here.
    # allowed_mentions: dict
    flags: int
    components: Optional[List[ext_components.Component]]

    def __init__(self):
        '''InteractionCallback.'''
        self.flags = 0
        self.embeds = list()
        self.components = list()

    def to_dict(self) -> dict:
        '''Convert object to dictionary suitable for API or other generic useage.'''
        new_dict: Dict[str, object] = dict()
        if type(self.flags) is enumerations.INTERACTION_CALLBACK_FLAGS:
            new_dict['flags'] = self.flags.value
        else:
            new_dict['flags'] = self.flags
        if hasattr(self, 'tts'):
            new_dict['tts'] = self.tts
        if hasattr(self, 'content'):
            new_dict['content'] = self.content
        if hasattr(self, 'components'):
            new_dict['components'] = list()
            assert type(self.components) is list
            assert type(new_dict['components']) is list
            for component in self.components:
                new_dict['components'].append(component.to_dict())
        if hasattr(self, 'embeds'):
            new_dict['embeds'] = list()
            assert type(self.embeds) is list
            assert type(new_dict['embeds']) is list
            for embed in self.embeds:
                new_dict['embeds'].append(embed.to_dict())
        return new_dict

    def generate(self,
                 tts: Optional[bool] = None,
                 content: Optional[str] = None,
                 ephemeral: Optional[bool] = None,
                 ):
        '''Customize common attributes of a callback.

        Arguments:
            tts (bool): Send message as text to speech.
            content (str): Text contents of the message to send.
            ephemeral (bool): Only show message to the target user. Not meaningful if message is a followup.
        '''
        if tts is not None:
            self.tts = tts
        if content is not None:
            self.content = content
        if ephemeral is True:
            self.flags |= self.INTERACTION_CALLBACK_FLAGS.EPHEMERAL
        elif ephemeral is False:
            self.flags &= ~self.INTERACTION_CALLBACK_FLAGS.EPHEMERAL
        self.components = []



from typing import Union, Optional
from abc import ABC

from .. import utilities
from ..client import api
from . import snowflake, message as ext_message, enumerations


class Channel(ABC):
    '''Abstract base class for Channels.'''

    _log = utilities.Log()

    CHANNEL_TYPES = enumerations.CHANNEL_TYPES

    id: 'snowflake.Snowflake' = None  # type: ignore
    name: str = None  # type: ignore
    position: int = None  # type: ignore
    type: 'enumerations.CHANNEL_TYPES' = None  # type: ignore

    def __str__(self):
        '''Return string representation.'''
        return f'{self.__class__.__name__}(name=\'{self.name}\', type={self.type.name})'

    def __repr__(self):
        '''Return string representation.'''
        return self.__str__()

    def from_dict(self, data, parent_guild=None) -> 'Channel':
        '''Parse a Channel from an API compliant dict.'''
        self.id = snowflake.Snowflake(data['id'])
        self.type = self.CHANNEL_TYPES(data['type'])
        self.name = data.get('name', '<DirectMessage>')
        return self

    @classmethod
    async def get_channel(cls, channel_id: snowflake.Snowflake):
        '''Invoke cache or API to get a channel of given channel_id.'''
        return ChannelImporter().from_dict(await api.API.get_channel(channel_id))


class TextChannel(Channel):
    '''Text channel. May be public or private depending on privacy settings.'''

    guild_id: 'snowflake.Snowflake' = None  # type: ignore
    # permission_overwrites: list
    rate_limit_per_user: int = None  # type: ignore
    nsfw: Optional[bool] = None  # type: ignore
    topic: str = None  # type: ignore
    last_message_id: 'snowflake.Snowflake' = None  # type: ignore
    parent_id: 'snowflake.Snowflake' = None  # type: ignore
    default_auto_archive_duration: int = None  # type: ignore

    def from_dict(self, data, parent_guild=None):
        '''Parse a TextChannel from an API compliant dict.'''
        self.id = snowflake.Snowflake(data['id'])
        if 'guild_id' in data:
            self.guild_id = snowflake.Snowflake(data['guild_id'])
        self.name = data['name']
        self.type = enumerations.CHANNEL_TYPES(data['type'])
        if 'position' in data:
            self.position = data['position']
        if 'rate_limit_per_user' in data:
            self.rate_limit_per_user = data['rate_limit_per_user']
        if 'nsfw' in data:
            self.nsfw = data['nsfw']
        if 'topic' in data:
            self.topic = data['topic']
        if 'last_message_id' in data:
            self.last_message_id = snowflake.Snowflake(data['last_message_id'])
        if 'parent_id' in data:
            self.parent_id = snowflake.Snowflake(data['parent_id'])
        if 'default_auto_archive_duration' in data:
            self.default_auto_archive_duration = data['default_auto_archive_duration']
        return self

    async def send_message(self, message: Union[ext_message.Message, str]) -> 'ext_message.Message':
        '''Send message to this channel. Will also accept a string.'''
        if type(message) is ext_message.Message:
            message.validate()
        elif type(message) is str:
            new_message = ext_message.Message()
            new_message.content = message
            message = new_message
        else:
            raise TypeError(f'send_message given invalid input of [{type(message)}].')

        data = await api.API.create_message(self.id, message.to_sendable_dict())
        return ext_message.Message().from_dict(data)


class NewsChannel(Channel):
    '''News channel.'''
    pass


class VoiceChannel(Channel):
    '''Voice channel.'''
    pass


class DMChannel(Channel):
    '''Direct Message with a specific user.'''
    pass


class GroupDMChannel(Channel):
    '''DM with several users.'''
    pass


class GuildPublicThread(Channel):
    '''Public thread.'''
    pass


class GuildPrivateThread(Channel):
    '''Private thread.'''
    pass


class CategoryChannel(Channel):
    '''Not a text channel, used for grouping servers in discord's GUI.'''
    pass


class StoreChannel(Channel):
    '''Channel to sell things in.'''
    pass


class ChannelImporter:
    '''Dynamic Channel identifier and parser.'''

    @classmethod
    def from_dict(cls, data, parent_guild=None) -> 'Union[TextChannel, DMChannel, VoiceChannel, GroupDMChannel, GuildPublicThread, GuildPrivateThread, CategoryChannel, NewsChannel, StoreChannel]':
        '''Parse a Channel from an API compliant dict, then return the appropriate subclass object.'''
        new_channel: Union[TextChannel,
                           DMChannel,
                           VoiceChannel,
                           GroupDMChannel,
                           GuildPublicThread,
                           GuildPrivateThread,
                           CategoryChannel,
                           NewsChannel,
                           StoreChannel,
                           ]

        if data["type"] == Channel.CHANNEL_TYPES.GUILD_TEXT:
            new_channel = TextChannel()

        elif data["type"] == Channel.CHANNEL_TYPES.DM:
            new_channel = DMChannel()

        elif data["type"] == Channel.CHANNEL_TYPES.GUILD_VOICE:
            new_channel = VoiceChannel()

        elif data["type"] == Channel.CHANNEL_TYPES.GROUP_DM:
            new_channel = GroupDMChannel()

        elif data["type"] == Channel.CHANNEL_TYPES.GUILD_PUBLIC_THREAD:
            new_channel = GuildPublicThread()

        elif data["type"] == Channel.CHANNEL_TYPES.GUILD_PRIVATE_THREAD:
            new_channel = GuildPrivateThread()

        elif data["type"] == Channel.CHANNEL_TYPES.GUILD_CATEGORY:
            new_channel = CategoryChannel()

        elif data["type"] == Channel.CHANNEL_TYPES.GUILD_NEWS:
            new_channel = NewsChannel()

        elif data["type"] == Channel.CHANNEL_TYPES.GUILD_STORE:
            new_channel = StoreChannel()

        else:
            raise ValueError(f'Dict contained unknown channel type. {data["type"]}, {Channel.CHANNEL_TYPES(data["type"]).name}')

        new_channel.from_dict(data, parent_guild)

        return new_channel



from abc import ABC
from ..utilities import log


class BaseDiscordObject(ABC):
    '''Abstract base of all common discord objects. All subclasses map directly to an actual API object.'''

    _log = log.Log()

    def from_dict(self, data: dict) -> 'BaseDiscordObject':
        '''Parse an object from a dictionary and return it.'''
        raise NotImplementedError(f'{self.__class__.__name__} does not yet implement this function.')

    def to_dict(self) -> dict:
        '''Convert object to dictionary suitable for API or other generic useage.'''
        raise NotImplementedError(f'{self.__class__.__name__} does not yet implement this function.')



'''Interactions with the discord API.

For more information, see [Discord API](https://discord.com/developers/docs/interactions/application-commands).
'''

from .command import Command, CommandOptions, CommandOptionChoiceStructure
from .interaction import Interaction, InteractionResponse, InteractionData
from .components import Component, ActionRow, Button, SelectMenu
from .enumerations import CHANNEL_TYPE, COMMAND_OPTION, COMMAND_TYPE, BUTTON_STYLES, INTERACTION_TYPES, INTERACTION_RESPONSE_TYPES,\
    INTERACTION_CALLBACK_FLAGS

__all__ = [
    'ActionRow',
    'BUTTON_STYLES',
    'Button',
    'CHANNEL_TYPE',
    'COMMAND_OPTION',
    'COMMAND_TYPE',
    'Command',
    'CommandOptionChoiceStructure',
    'CommandOptions',
    'Component',
    'INTERACTION_CALLBACK_FLAGS',
    'INTERACTION_RESPONSE_TYPES',
    'INTERACTION_TYPES',
    'Interaction',
    'InteractionData',
    'InteractionResponse',
    'SelectMenu',
]
